# -*- coding: utf-8 -*-

import json
import shutil
from pathlib import Path

from micropy import data, utils
from micropy.exceptions import StubError, StubValidationError
from micropy.logger import Log
from micropy.stubs import source


class StubManager:
    """Manages a collection of Stubs.

    Kwargs:
        resource (str): Default resource path
        repos ([StubRepo]): Repos for Remote Stubs

    Raises:
        StubError: a stub is missing a def file
        StubValidationError: a stubs def file is not valid

    Returns:
        object: Instance of StubManager

    """

    _schema = data.SCHEMAS / "stubs.json"
    _firm_schema = data.SCHEMAS / "firmware.json"

    def __init__(self, resource=None, repos=None):
        self._loaded = set()
        self._firmware = set()
        self.resource = resource
        self.repos = repos
        self.log = Log.add_logger("Stubs", stdout=False, show_title=False)
        if self.resource:
            self.load_from(resource, strict=False)

    def __iter__(self):
        return iter(self._loaded)

    def __len__(self):
        return len(self._loaded)

    def iter_by_firmware(self, stubs=None):
        """Iterate stubs sorted by firmware.

        Args:
            stubs ([Stub], optional): Sublist of Stubs to iterate over.
                Defaults to None. If none, uses all installed stubs.

        """
        loaded = stubs or self._loaded
        for firm in self._firmware:
            stubs = [s for s in loaded if s.firmware == firm]
            yield (firm, stubs)
        other = [s for s in loaded if s.firmware is None]
        yield ("Unknown", other)

    def verbose_log(self, state):
        """Enable Stub logging to stdout.

        Args:
            state (bool): State to set

        Returns:
            bool: state

        """
        self.log.stdout = state
        return state

    def _load(self, stub_source, strict=True, **kwargs):
        """Loads a stub into StubManager.

        Args:
            stub_source (StubSource): Stub Source Instance
            strict (bool, optional): Raise Exception if stub fails to resolve.
                Defaults to True.

        Raises:
            e: Exception raised by resolving failure

        Returns:
            Stub: Instance of Stub

        """
        with stub_source.ready() as src_path:
            try:
                stub_type = self._get_stubtype(src_path)
            except Exception as e:
                self.log.debug(f"{src_path.name} failed to validate: {e}")
                if strict:
                    raise e
            else:
                if stub_type is FirmwareStub:
                    fware = stub_type(src_path, **kwargs)
                    self._firmware.add(fware)
                    self.log.debug(f"Firmware Loaded: {fware}")
                    return fware
                stub = stub_type(src_path, **kwargs)
                fware = self.resolve_firmware(stub)
                stub.firmware = fware
                self._loaded.add(stub)
                self.log.success(f"{stub.name} added!")
                self.log.debug(f"Loaded: {stub}")
                return stub

    def resolve_firmware(self, stub):
        """Resolves FirmwareStub for DeviceStub instance.

        Args:
            stub (DeviceStub): Stub to resolve

        Returns:
            FirmwareStub: Instance of FirmwareStub
            NoneType: None if an appropriate
                FirmwareStub cannot be found

        """
        fware_name = stub.firmware_name
        self.log.info(f"Detected Firmware: $[{fware_name}]")
        results = (f for f in self._firmware if f.firmware == fware_name)
        fware = next(results, None)
        if not fware:
            try:
                self.log.info("Firmware not found locally, attempting to install it...")
                fware = self.add(fware_name)
            except Exception:
                self.log.error("Failed to resolve firmware!")
                return None
            else:
                self.log.success(f"{fware_name} firmware added!")
                return fware
        return fware

    def validate(self, path, schema=None):
        """Validates given stub path against its schema.

        Args:
            path (str): path to validate
            schema (str, optional): Path to schema. Defaults to None.
                If None, the DeviceStub schema is used.

        Raises:
            StubError: Raised if no info file can be found
            StubValidationError: Raised if the info file fails validation

        """
        self.log.debug(f"Validating: {path}")
        schema = schema or self._schema
        path = Path(path).resolve()
        stub_info = path / "info.json"
        val = utils.Validator(schema)
        try:
            val.validate(stub_info)
        except FileNotFoundError:
            raise StubError(f"{path.name} contains no info file!")
        except Exception as e:
            raise StubValidationError(path, str(e))

    def _get_stubtype(self, path):
        """Resolves appropriate stub type.

        Args:
            path (str): path to stub

        Returns:
            cls: Appropriate class for stub

        """
        try:
            self.validate(path)
        except StubValidationError:
            try:
                self.validate(path, schema=self._firm_schema)
            except Exception as e:
                raise e
            else:
                return FirmwareStub
        except Exception as e:
            raise e
        else:
            return DeviceStub

    def is_valid(self, path):
        """Check if stub is valid without raising an exception.

        Args:
            path (str): path to stub

        Returns:
            bool: True if stub is valid

        """
        try:
            self._get_stubtype(path)
        except Exception:
            return False
        else:
            return True

    def _check_existing(self, location):
        """check if location is or contains an existing stub.

        Args:
            location (str): name or path of Stub

        Returns:
            generator of existing stubs

        """
        try:
            do_recurse = self._should_recurse(location)
        except StubError:
            yield
        else:
            if do_recurse:
                for s in (self._check_existing(p) for p in location.iterdir()):
                    yield next(s, None)
            path_name = Path(location).name
            stub = next(
                (
                    s
                    for s in self._loaded
                    if any(t in (s.name, s.path.name) for t in (path_name, location))
                ),
                None,
            )
            if stub:
                yield stub

    def load_from(self, directory, *args, **kwargs):
        """Recursively loads stubs from a directory.

        Args:
            directory (str): Path to load from

        Returns:
            [DeviceStub]: List of loaded Stubs

        """
        dir_path = Path(str(directory)).resolve()
        dirs = dir_path.iterdir()
        sources = [source.get_source(d) for d in dirs]
        stubs = []
        for stub in sources.copy():
            if self.is_valid(stub.location):
                stub_type = self._get_stubtype(stub.location)
                if stub_type is FirmwareStub:
                    sources.remove(stub)
                    self._load(stub, *args, **kwargs)
        stubs.extend([self._load(s, *args, **kwargs) for s in sources])
        return stubs

    def _should_recurse(self, location):
        """Checks for multiple stubs in a location.

        Args:
            location (str): location of potential stub

        Raises:
            StubError: No info files could be found

        Returns:
            bool: True if multiple stubs are found

        """
        if not Path(location).exists():
            return False
        path = Path(location).resolve()
        info_glob = list(path.rglob("info.json"))
        if len(info_glob) == 0:
            raise StubError(f"{path.name} contains no info file!")
        if len(info_glob) > 1:
            return True
        return False

    def add(self, location, dest=None, force=False):
        """Add stub(s) from source.

        Args:
            source (str): path to stub(s)
            dest (str, optional): path to copy stubs to.
                Defaults to self.resource
            force (bool, optional): overwrite existing stubs.
                Defaults to False.

        Raises:
            TypeError: No resource or destination provided

        """
        _dest = dest or self.resource
        if not _dest:
            raise TypeError("No Stub Destination Provided!")
        dest = Path(str(_dest)).resolve()
        stubs = [s for s in self._check_existing(location) if s is not None]
        if any(stubs):
            for stub in stubs:
                if not force:
                    self.log.info(f"$[{stub}] is already installed!")
                    return stub
                self.log.info(f"Uninstalling $[{stub.name}]...")
                shutil.rmtree(stub.path)
        if self._should_recurse(location):
            return self.load_from(location, strict=False, copy_to=dest)
        self.log.info(f"\nResolving stub...")
        stub_source = source.get_source(location, log=self.log)
        return self._load(stub_source, copy_to=dest)

    def from_stubber(self, path, dest):
        """Formats stubs generated by createstubs.py.

        Creates a stub package from the stubs generated by
        createstubs.py. Also attempts to auto-resolve the stubs
        firmware name.

        Args:
            path (str): path to generated stubs
            dest (str): path to output

        Returns:
            str: formatted stubs

        """
        _path = Path(path).resolve()
        dest = Path(dest).resolve()
        mod_file = next(_path.rglob("modules.json"))
        path = mod_file.parent
        mod_data = json.load(mod_file.open())
        dev_fware = mod_data["firmware"]
        fname = dev_fware.get("name", None)
        out_name = f"{dev_fware['sysname']}"
        # TODO: Attempt to Autoresolve Firmware name and add it to info.json
        if fname:
            out_name = f"{out_name}-{fname}"
        out_name = f"{out_name}-{dev_fware['version']}"
        out_stub = dest / out_name
        info_file = out_stub / "info.json"
        stub_path = out_stub / "stubs"
        out_stub.mkdir(exist_ok=True, parents=True)
        json.dump(mod_data, info_file.open("w+"))
        shutil.copytree(path, stub_path)
        return out_stub

    def search_remote(self, query):
        """Search all repositories for query.

        Args:
            query (str): query to search for

        Returns:
            [tuple]: List of result tuples. The first item
                is the package name, and the second is a bool
                based on whether the package is installed or not

        """
        results = []
        installed = [str(s) for s in self._loaded.union(self._firmware)]
        for repo in self.repos:
            for p in repo.search(query):
                results.append((p, p in installed))
        return sorted(results)

    def resolve_subresource(self, stubs, subresource):
        """Resolve or Create StubManager from list of stubs.

        Args:
            stubs ([Stub]): List of stubs to use in subresource
            subresource (str): path to subresource

        Returns:
            StubManager: StubManager with subresource stubs

        """
        for stub in stubs:
            fware = stub.firmware
            if fware:
                link = subresource / fware.path.name
                fware = FirmwareStub.resolve_link(fware, link)
            link = subresource / stub.path.name
            stub = DeviceStub.resolve_link(stub, link)
            stub.firmware = fware
            yield stub


class Stub:
    """Abstract Parent for Stub Related Classes.

    Not Meant to be instantiated directly. Contains common logic
    between different types of Stubs (ex. Firmware vs Device)

    Raises:
        NotImplementedError: name property is not implemented

    Returns:
        Instance of Stub

    """

    def __init__(self, path, copy_to=None, **kwargs):
        self.path = Path(path)
        ref = self.path / "info.json"
        self.info = json.load(ref.open())
        if copy_to is not None:
            self.copy_to(copy_to)

    def copy_to(self, dest, name=None):
        """Copy stub to a directory."""
        if not name:
            dest = Path(dest) / self.path.name
        shutil.copytree(self.path, dest)
        self.path = dest.resolve()
        return self

    @classmethod
    def resolve_link(cls, stub, link_path):
        """Resolve or Create Stub Symlink.

        Args:
            stub (Stub): stub to resolve
            link_path (str): path to link

        Returns:
            Stub: Stub from symlink

        """
        fware = stub.firmware
        if utils.is_dir_link(link_path):
            return cls(link_path, firmware=fware)
        utils.create_dir_link(link_path, stub.path)
        return cls(link_path, firmware=fware)

    @property
    def name(self):
        """Human friendly stub name."""
        raise NotImplementedError

    def __eq__(self, other):
        return self.name == getattr(other, "name", None)

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name


class DeviceStub(Stub):
    """Handles Device Specific Stubs.

    Args:
        path (str): path to stub
        copy_to (str, optional): Path to copy Stub to. Defaults to None.

    Returns:
        Device Stub Instance

    """

    def __init__(self, path, copy_to=None, **kwargs):
        super().__init__(path, copy_to, **kwargs)

        self.stubs = self.path / "stubs"
        self.frozen = self.path / "frozen"
        stubber = self.info.get("stubber")
        self.stub_version = stubber.get("version")

        self.firm_info = self.info.get("firmware")
        self.firmware = kwargs.get("firmware", None)
        self.sysname = self.firm_info.get("sysname")
        self.version = self.firm_info.get("version")

    @property
    def firmware_name(self):
        """Return an appropriate firmware name.

        Returns:
            str: Name of Firmware

        """
        if isinstance(self.firmware, FirmwareStub):
            return self.firmware.firmware
        fware = self.firm_info.get("name", None)
        if not fware:
            fware = self.firm_info.get("firmware").strip()
            fware.replace(" ", "-")
        return fware

    @property
    def name(self):
        if not isinstance(self.firmware, FirmwareStub):
            return f"{self.sysname}-{self.version}"
        return f"{self.sysname}-{self.firmware_name}-{self.version}"

    def __repr__(self):
        return (
            f"DeviceStub(sysname={self.sysname}, firmware="
            f"{self.firmware_name}, version={self.version}, "
            f"path={self.path})"
        )


class FirmwareStub(Stub):
    """Handles Firmware Specific Modules.

    Args:
        path (str): path to stub
        copy_to (str, optional): Path to copy Stub to. Defaults to None.

    Returns:
        FirmwareStub Instance

    """

    def __init__(self, path, copy_to=None, **kwargs):
        super().__init__(path, copy_to=copy_to, **kwargs)

        self.frozen = self.path / "frozen"
        self.repo = self.info.get("repo")
        firmware = self.info.get("firmware").strip()
        self.firmware = firmware.replace(" ", "-")

    @property
    def name(self):
        return self.firmware

    def __repr__(self):
        return f"FirmwareStub(firmware={self.firmware}, repo={self.repo})"



# -*- coding: utf-8 -*-

"""Configuration files and interfaces for them."""

from .config import Config
from .config_dict import DictConfigSource
from .config_json import JSONConfigSource

__all__ = ["Config", "JSONConfigSource", "DictConfigSource"]



f='esp32'
e='{}/{}'
d='logging'
c='sys'
b='upip'
a=IndexError
Z=NameError
Y=print
X=NotImplementedError
S='machine'
R=True
Q='_thread'
P='__init__'
N='version'
M=len
L=KeyError
K=ImportError
J='.'
I='_'
G=False
H=''
F=AttributeError
E='/'
D=None
C=OSError
import sys,gc as B,uos as os
from utime import sleep_us as g
from ujson import dumps as O
__version__='1.4.3'
T=2
h=2
try:from machine import resetWDT as U
except K:
	def U():0
class Stubber:
	def __init__(A,path=D,firmware_id=D):
		G=firmware_id;D=path
		try:
			if os.uname().release=='1.13.0'and os.uname().version<'v1.13-103':raise X('MicroPython 1.13.0 cannot be stubbed')
		except F:pass
		A._report=[];A.info=_info();B.collect()
		if G:A._fwid=str(G).lower()
		else:A._fwid='{family}-{port}-{ver}'.format(**A.info).lower()
		A._start_free=B.mem_free()
		if D:
			if D.endswith(E):D=D[:-1]
		else:D=get_root()
		A.path='{}/stubs/{}'.format(D,A.flat_fwid).replace('//',E)
		try:V(D+E)
		except C:pass
		A.problematic=[b,'upysh','webrepl_setup','http_client','http_client_ssl','http_server','http_server_ssl'];A.excluded=['webrepl','_webrepl','port_diag','example_sub_led.py','example_pub_button.py'];A.modules=[]
	def get_obj_attributes(K,item_instance):
		J="Couldn't get attribute '{}' from object '{}', Err: {}";A=item_instance;C=[];E=[]
		try:
			for D in dir(A):
				try:G=getattr(A,D);C.append((D,repr(G),repr(type(G)),G))
				except F as H:E.append(J.format(D,A,H))
		except F as H:E.append(J.format(D,A,H))
		C=[A for A in C if not(A[0].startswith(I)and A[0]!=P)];B.collect();return C,E
	def add_modules(A,modules):A.modules=sorted(set(A.modules)|set(modules))
	def create_all_stubs(A):
		B.collect()
		for C in A.modules:A.create_one_stub(C)
	def create_one_stub(D,module_name):
		A=module_name
		if A.startswith(I)and A!=Q:return G
		if A in D.problematic:return G
		if A in D.excluded:return G
		F='{}/{}.py'.format(D.path,A.replace(J,E));B.collect();H=B.mem_free();Y('Stub module: {:<20} to file: {:<55} mem:{:>5}'.format(A,F,H))
		try:D.create_module_stub(A,F)
		except C:return G
		B.collect();return R
	def create_module_stub(F,module_name,file_name=D):
		G=file_name;A=module_name
		if A.startswith(I)and A!=Q:return
		if A in F.problematic:return
		if G is D:G=F.path+E+A.replace(J,I)+'.py'
		if E in A:A=A.replace(E,J)
		M=D
		try:M=__import__(A,D,D,'*')
		except K:return
		V(G)
		with open(G,'w')as N:O='"""\nModule: \'{0}\' on {1}\n"""\n# MCU: {2}\n# Stubber: {3}\n'.format(A,F._fwid,F.info,__version__);N.write(O);N.write('from typing import Any\n\n');F.write_object_stub(N,M,A,H)
		F._report.append({'module':A,'file':G})
		if not A in['os',c,d,'gc']:
			try:del M
			except (C,L):pass
			try:del sys.modules[A]
			except L:pass
		B.collect()
	def write_object_stub(K,fp,object_expr,obj_name,indent,in_class=0):
		b='tuple';a='list';Y='dict';X='{0}{1} = {2} # type: {3}\n';W='bound_method';V='Any';Q=in_class;O=object_expr;I=fp;A=indent;B.collect()
		if O in K.problematic:return
		R,c=K.get_obj_attributes(O)
		for (D,J,F,S) in R:
			if D in['classmethod','staticmethod']:continue
			U();g(1)
			if F=="<class 'type'>"and M(A)<=h*4:E='\n'+A+'class '+D+':\n';E+=A+"    ''\n";I.write(E);K.write_object_stub(I,S,'{0}.{1}'.format(obj_name,D),A+'    ',Q+1)
			elif'method'in F or'function'in F or D==P:
				N=V;T=H
				if Q>0:
					T='self, '
					if D==P:N='None'
				if W in F or W in J:E='{}@classmethod\n'.format(A);E+='{}def {}(cls, *args) -> {}:\n'.format(A,D,N)
				else:E='{}def {}({}*args) -> {}:\n'.format(A,D,T,N)
				E+=A+'    ...\n\n';I.write(E)
			elif F=="<class 'module'>":0
			elif F.startswith("<class '"):
				G=F[8:-2];E=H
				if G in['str','int','float','bool','bytearray','bytes']:E=X.format(A,D,J,G)
				elif G in[Y,a,b]:d={Y:'{}',a:'[]',b:'()'};E=X.format(A,D,d[G],G)
				else:
					if not G in['object','set','frozenset']:G=V
					E='{0}{1} : {2} ## {3} = {4}\n'.format(A,D,G,F,J)
				I.write(E)
			else:I.write("# all other, type = '{0}'\n".format(F));I.write(A+D+' # type: Any\n')
		del R;del c
		try:del D,J,F,S
		except (C,L,Z):pass
	@property
	def flat_fwid(self):
		A=self._fwid;B=' .()/\\:$'
		for C in B:A=A.replace(C,I)
		return A
	def clean(E,path=D):
		A=path
		if A is D:A=E.path
		Y('Clean/remove files in folder: {}'.format(A))
		try:G=os.listdir(A)
		except (C,F):return
		for H in G:
			B=e.format(A,H)
			try:os.remove(B)
			except C:
				try:E.clean(B);os.rmdir(B)
				except C:pass
	def report(D,filename='modules.json'):
		E=',\n';H=e.format(D.path,filename);B.collect()
		try:
			with open(H,'w')as A:
				A.write('{');A.write(O({'firmware':D.info})[1:-1]);A.write(E);A.write(O({'stubber':{N:__version__}})[1:-1]);A.write(E);A.write('"modules" :[\n');F=R
				for I in D._report:
					if F:F=G
					else:A.write(E)
					A.write(O(I))
				A.write('\n]}')
			J=D._start_free-B.mem_free()
		except C:pass
def V(path):
	B=path;A=F=0
	while A!=-1:
		A=B.find(E,F)
		if A!=-1:
			if A==0:D=B[0]
			else:D=B[0:A]
			try:I=os.stat(D)
			except C as G:
				if G.args[0]==T:
					try:os.mkdir(D)
					except C as H:raise H
				else:raise G
		F=A+1
def _info():
	h='loboris';g=' on ';e='0.0.0';d='port';c='platform';b='nodename';Z='name';U='v';T='mpy';R='unknown';Q='-';P='sysname';I='ver';G='family';E='build';B='release';V=sys.implementation.name;W=sys.platform;A={Z:V,B:e,N:e,E:H,P:R,b:R,S:R,G:V,c:W,d:W,I:H}
	try:A[B]=J.join([str(A)for A in sys.implementation.version]);A[N]=A[B];A[Z]=sys.implementation.name;A[T]=sys.implementation.mpy
	except F:pass
	if sys.platform not in('unix','win32'):
		try:
			C=os.uname();A[P]=C.sysname;A[b]=C.nodename;A[B]=C.release;A[S]=C.machine
			if g in C.version:
				O=C.version.split(g)[0]
				if A[P]=='esp8266':
					if Q in O:X=O.split(Q)[0]
					else:X=O
					A[N]=A[B]=X.lstrip(U)
				try:A[E]=O.split(Q)[1]
				except a:pass
		except (a,F,TypeError):pass
	try:from pycopy import const;A[G]='pycopy';del const
	except (K,L):pass
	if A[c]=='esp32_LoBo':A[G]=h;A[d]=f
	elif A[P]=='ev3':
		A[G]='ev3-pybricks';A[B]='1.0.0'
		try:from pybricks.hubs import EV3Brick;A[B]='2.0.0'
		except K:pass
	if A[B]:A[I]=U+A[B].lstrip(U)
	if A[G]!=h:
		if A[B]and A[B]>='1.10.0'and A[B].endswith('.0'):A[I]=A[B][:-2]
		else:A[I]=A[B]
		if A[E]!=H and M(A[E])<4:A[I]+=Q+A[E]
	if T in A:
		i=int(A[T]);Y=[D,'x86','x64','armv6','armv6m','armv7m','armv7em','armv7emsp','armv7emdp','xtensa','xtensawin'][i>>10]
		if Y:A['arch']=Y
	return A
def get_root():
	A='/flash'
	try:D=os.stat(A)
	except C as B:
		if B.args[0]==T:
			try:A=os.getcwd()
			except (C,F):A=J
		else:A=E
	return A
def A():sys.exit(1)
def read_path():
	B=H
	if M(sys.argv)==3:
		C=sys.argv[1].lower()
		if C in('--path','-p'):B=sys.argv[2]
		else:A()
	elif M(sys.argv)==2:A()
	return B
def W():
	try:A=bytes('abc',encoding='utf8');B=W.__module__;return G
	except (X,F):return R
def main():stubber=Stubber(path=read_path());stubber.clean();stubber.modules=['_onewire',Q,'_uasyncio','ak8963','apa102','apa106','array','binascii','btree','builtins','cmath','collections','crypto','curl','dht','display','display_driver_utils','ds18x20','errno','esp',f,'espidf','flashbdev','framebuf','freesans20','fs_driver','functools','gc','gsm','hashlib','heapq','ili9XXX','imagetools','inisetup','io','json','lcd160cr','lodepng',d,'lv_colors','lv_utils','lvgl','lwip',S,'math','microWebSocket','microWebSrv','microWebTemplate','micropython','mpu6500','mpu9250','neopixel','network','ntptime','onewire','os','pyb','pycom','pye','queue','random','re','requests','rtch','select','socket','ssd1306','ssh','ssl','stm','struct',c,'time','tpcalib','uarray','uasyncio/__init__','uasyncio/core','uasyncio/event','uasyncio/funcs','uasyncio/lock','uasyncio/stream','ubinascii','ubluetooth','ucollections','ucrypto','ucryptolib','uctypes','uerrno','uftpd','uhashlib','uheapq','uio','ujson','ulab','ulab/approx','ulab/compare','ulab/fft','ulab/filter','ulab/linalg','ulab/numerical','ulab/poly','ulab/user','ulab/vector','umachine','umqtt/robust','umqtt/simple','uos',b,'upip_utarfile','uqueue','urandom','ure','urequests','urllib/urequest','uselect','usocket','ussl','ustruct','usys','utelnetserver','utime','utimeq','uwebsocket','uzlib','websocket','websocket_helper','writer','xpt2046','ymodem','zlib'];B.collect();stubber.create_all_stubs();stubber.report()
if __name__=='__main__'or W():
	try:logging.basicConfig(level=logging.INFO)
	except Z:pass
	main()


"""
Module: 'micropython' on esp32 1.11.0
"""
# MCU: (sysname='esp32', nodename='esp32', release='1.11.0', version='v1.11-132-gc24d81119 on 2019-07-08', machine='ESP32 module with ESP32')
# Stubber: 1.2.0
def alloc_emergency_exception_buf():
    pass


def const():
    pass


def heap_lock():
    pass


def heap_unlock():
    pass


def kbd_intr():
    pass


def mem_info():
    pass


def opt_level():
    pass


def qstr_info():
    pass


def schedule():
    pass


def stack_use():
    pass



from __future__ import annotations

from pathlib import Path
from typing import AnyStr, Type

from .abc import (
    DevicePath,
    HostPath,
    MessageConsumer,
    MetaPyDevice,
    MetaPyDeviceBackend,
    StreamConsumer,
)
from .backend_upydevice import UPyDeviceBackend
from .consumers import ConsumerDelegate


class PyDevice(MetaPyDevice):
    pydevice: MetaPyDeviceBackend
    consumer: ConsumerDelegate

    def __init__(
        self,
        location: str,
        *,
        backend: Type[MetaPyDeviceBackend] = UPyDeviceBackend,
        auto_connect: bool = True,
        stream_consumer: StreamConsumer = None,
        message_consumer: MessageConsumer = None,
        delegate_cls: Type[ConsumerDelegate] = ConsumerDelegate,
    ):
        self.pydevice = backend().establish(location)
        self.consumer = delegate_cls(stream_consumer, message_consumer)
        if auto_connect and self.pydevice:
            self.pydevice.connect()

    def copy_from(self, source_path: DevicePath, target_path: HostPath) -> None:
        src_path = Path(str(source_path))
        # 'is_dir/file' only works on existing paths.
        if not src_path.suffix:
            return self.pydevice.copy_dir(
                DevicePath(source_path), target_path, consumer=self.consumer
            )
        return self.pydevice.pull_file(DevicePath(source_path), target_path, consumer=self.consumer)

    def copy_to(self, source_path: HostPath, target_path: DevicePath) -> None:
        src_path = Path(str(source_path))
        host_exists = src_path.exists()
        if (host_exists and src_path.is_dir()) or (not host_exists and not src_path.suffix):
            raise RuntimeError("Copying dirs to device is not yet supported!")
        return self.pydevice.push_file(source_path, target_path, consumer=self.consumer)

    def connect(self):
        return self.pydevice.connect()

    def disconnect(self):
        return self.pydevice.disconnect()

    def run_script(self, content: AnyStr, target_path: DevicePath | None = None):
        return self.pydevice.eval_script(content, target_path, consumer=self.consumer)



# -*- coding: utf-8 -*-

from pathlib import Path

import pytest
from boltons.namedutils import namedlist
from micropy import packages

EXPPKG = namedlist("ExpectPackage", ["name", "specs", "full_name"])


class TestPackages:
    class MockSource:
        def __init__(self, pkg, has_init):
            self.pkg = pkg
            self.has_init = has_init

    @pytest.fixture(params=[True, False], ids=["package", "module"])
    def mock_source(self, request, mock_pkg):
        if request.param:
            # true packages vs file
            (mock_pkg / "__init__.py").touch()
        return self.MockSource(mock_pkg, request.param)

    @pytest.fixture(params=[True, False], ids=["package", "module"])
    def mock_source_path(self, request, tmp_path):
        path = tmp_path / "file.py"
        if request.param is True:
            pkg = tmp_path
            path = tmp_path / "__init__.py"
        else:
            pkg = tmp_path
        path.touch()
        return self.MockSource(pkg, request.param)

    @pytest.mark.parametrize(
        "package,expect",
        [
            (
                "git+https://github.com/jczic/MicroWebSrv2.git@master#egg=MicroWebSrv2",
                packages.VCSDependencySource,
            ),
            ("picoweb", packages.PackageDependencySource),
            ("-e /foobar/pkg", packages.LocalDependencySource),
        ],
    )
    def test_factory(self, package, expect):
        source = packages.create_dependency_source(package)
        assert isinstance(source, expect)

    @pytest.mark.parametrize(
        "requirement,expect",
        [
            (
                ["git+https://github.com/jczic/MicroWebSrv2.git@master#egg=MicroWebSrv2"],
                [
                    "microwebsrv2",
                    "git+https://github.com/jczic/MicroWebSrv2.git@master#egg=MicroWebSrv2",
                    "git+https://github.com/jczic/MicroWebSrv2.git@master#egg=MicroWebSrv2",
                ],
            ),
            (["picoweb"], ["picoweb", "*", "picoweb"]),
            (["picoweb==^7.1"], ["picoweb", "==^7.1", "picoweb==^7.1"]),
            (["BlynkLib==0.0.0"], ["blynklib", "==0.0.0", "blynklib==0.0.0"]),
            (
                ["-e /foobar/somepkg", "somepackage"],
                ["somepackage", "-e /foobar/somepkg", "-e /foobar/somepkg"],
            ),
            (["-e /foobar/somepkg"], ["somepkg", "-e /foobar/somepkg", "-e /foobar/somepkg"]),
        ],
    )
    def test_package(self, mock_pkg, requirement, expect):
        source = packages.create_dependency_source(*requirement)
        pkg = source.package
        assert pkg.name == expect[0]  # name
        assert pkg.pretty_specs == expect[1]  # specs
        assert str(pkg) == expect[2]  # full name ()

    def test_package_source(self, mock_source):
        def format_desc(x):
            return f"desc{x}"

        source = packages.create_dependency_source("blynklib", format_desc=format_desc)
        with source as files:
            if mock_source.has_init:
                assert isinstance(files, Path)
                assert mock_source.pkg.name == files.name
            else:
                assert isinstance(files, list)
                file_names = [(p.name, s.name) for p, s in files]
                file_names = list(sum(file_names, ()))
                assert sorted(["file.py", "file.pyi", "module.py", "module.pyi"]) == sorted(
                    file_names
                )

    def test_package_path(self, mock_source_path):
        source = packages.create_dependency_source(f"-e {mock_source_path.pkg}")
        with source as files:
            if mock_source_path.has_init:
                # is proper package
                assert files.is_dir()
            if mock_source_path.has_init is False:
                # is module
                assert files.is_dir()
                assert files == source.package.path

    @pytest.mark.parametrize(
        "pkg,expect",
        [
            (("micropy-cli", "*"), EXPPKG("micropy-cli", "*", "micropy-cli")),
            (("blynklib", "==0.0.0"), EXPPKG("blynklib", "==0.0.0", "blynklib==0.0.0")),
            (
                ("custompkg", "-e src/lib/custompackage"),
                EXPPKG("custompkg", "-e src/lib/custompackage", "-e src/lib/custompackage"),
            ),
        ],
    )
    def test_package_from_text(self, pkg, expect, utils):
        pkg = packages.Package.from_text(*pkg)
        assert pkg.name == expect.name
        assert pkg.full_name == pkg.full_name
        assert pkg.pretty_specs == expect.specs
        is_e = "-e" in pkg.full_name
        assert pkg.editable == is_e
        if pkg.editable:
            assert pkg.path == Path("src/lib/custompackage")
        else:
            assert pkg.path == None



# make_stub_files: Fri 21 Jun 2019 at 00:44:00

from typing import Any, Dict, Optional, Sequence, Tuple, Union

Node = Any
def time() -> Any: ...
    #   0: return val-NTP_DELTA
    # ? 0: return val-NTP_DELTA
def settime() -> None: ...



import os
import json
import logging
import shutil

from fnmatch import fnmatch
from pathlib import Path
from version import VERSION
from typing import List

import mypy.stubgen as stubgen
from mypy.errors import CompileError
import sys


log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

STUB_FOLDER = "./all-stubs"


def clean_version(version: str, build: bool = False):
    "omit the commit hash from the git tag"
    # 'v1.13-103-gb137d064e' --> 'v1.13-103'
    nibbles = version.split("-")
    if len(nibbles) == 1:
        return version
    elif build and build != "dirty":
        return "-".join(version.split("-")[0:-1])
    else:
        return "-".join((version.split("-")[0], "N"))


def stubfolder(path: str) -> str:
    "return path in the stub folder"
    return "{}/{}".format(STUB_FOLDER, path)


def flat_version(version: str, keep_v: bool = False):
    "Turn version from 'v1.2.3' into '1_2_3' to be used in filename"
    version = version.strip().replace(".", "_")
    if not keep_v:
        version = version.lstrip("v")
    return version


def cleanup(modules_folder: Path, all_pyi: bool = False):
    "Q&D cleanup"
    # for some reason (?) the umqtt simple.pyi and robust.pyi are created twice
    #  - modules_root folder ( simple.pyi and robust.pyi) - NOT OK
    #       - umqtt folder (simple.py & pyi and robust.py & pyi) OK
    #  similar for mpy 1.9x - 1.11
    #       - core.pyi          - uasyncio\core.py'
    #       - urequests.pyi     - urllib\urequest.py'
    # Mpy 1.13+
    #       - uasyncio.pyi      -uasyncio\__init__.py

    if all_pyi:
        for file in modules_folder.rglob("*.pyi"):
            file.unlink()
        # no need to remove anything else
        return

    for file_name in (
        "simple.pyi",
        "robust.pyi",
        "core.pyi",
        "urequest.pyi",
        "uasyncio.pyi",
    ):
        f = Path.joinpath(modules_folder, file_name)
        if f.exists():
            try:
                print(" - removing {}".format(f))
                f.unlink()
            except OSError:
                log.error(" * Unable to remove extranous stub {}".format(f))


def generate_pyi_from_file(file: Path) -> bool:
    """Generate a .pyi stubfile from a single .py module using mypy/stubgen"""

    sg_opt = stubgen.Options(
        pyversion=(3, 5),
        no_import=False,
        include_private=True,
        doc_dir="",
        search_path=[],
        interpreter=sys.executable,
        parse_only=False,
        ignore_errors=True,
        modules=[],
        packages=[],
        files=[],
        output_dir="",
        verbose=True,
        quiet=False,
        export_less=False,
    )
    # Deal with generator passed in
    if not isinstance(file, Path):
        raise TypeError
    sg_opt.files = [str(file)]
    sg_opt.output_dir = str(file.parent)
    try:
        print(f"Calling stubgen on {str(file)}")
        stubgen.generate_stubs(sg_opt)
        return True
    except (Exception, CompileError, SystemExit) as e:
        print(e)
        return False


def generate_pyi_files(modules_folder: Path) -> bool:
    """generate typeshed files for all scripts in a folder using mypy/stubgen"""
    # stubgen cannot process folders with duplicate modules ( ie v1.14 and v1.15 )

    modlist = list(modules_folder.glob("**/modules.json"))
    if len(modlist) > 1:
        # try to process each module seperatlely
        r = True
        for mod_manifest in modlist:
            ## generate fyi files for folder
            r = r and generate_pyi_files(mod_manifest.parent)
        return r
    else:  # one or less module manifests
        ## generate fyi files for folder
        # clean before to clean any old stuff
        # TODO: CLean out the old pyi files
        cleanup(modules_folder, all_pyi=True)

        print("running stubgen on {0}".format(modules_folder))
        cmd = "stubgen {0} --output {0} --include-private --ignore-errors".format(modules_folder)
        result = os.system(cmd)
        # Check on error
        if result != 0:
            # in case of failure ( duplicate module in subfolder) then Plan B
            # - run stubgen on each *.py
            print("::group::[stubgen] Failure on folder, attempt to run stubgen per file")
            py_files = list(modules_folder.rglob("*.py"))
            for py in py_files:
                generate_pyi_from_file(py)
                # todo: report failures by adding to module manifest

        # for py missing pyi:
        py_files = list(modules_folder.rglob("*.py"))
        pyi_files = list(modules_folder.rglob("*.pyi"))

        worklist = pyi_files.copy()
        for pyi in worklist:
            # remove all py files that have been stubbed successfully from the list
            try:
                py_files.remove(pyi.with_suffix(".py"))
                pyi_files.remove(pyi)
            except ValueError:
                log.info(f"no matching py for : {str(pyi)}")

        # if there are any pyi files remaining,
        # try to match them to py files and move them to the correct location
        worklist = pyi_files.copy()
        for pyi in worklist:
            match = [py for py in py_files if py.stem == pyi.stem]
            if match:
                # move the .pyi next to the corresponding .py
                log.info(f"moving : {str(pyi)}")
                src = str(pyi)
                dst = str(match[0].with_suffix(".pyi"))
                shutil.move(src, dst)
                try:
                    py_files.remove(match[0])
                    pyi_files.remove(pyi)
                except ValueError:
                    pass

        # now stub the rest
        # note in some cases this will try a file twice
        for py in py_files:
            generate_pyi_from_file(py)
            # todo: report failures by adding to module manifest

        # and clean after to only check-in good stuff
        # cleanup(modules_folder)
        return True
    #     ##
    # for mod_manifest in modlist:
    #     ## generate fyi files for folder
    #     generate_pyi_files(mod_manifest.parent)

    #     # todo: collect and report results
    # return True


def manifest(
    family=None,
    machine=None,
    port=None,
    platform=None,
    sysname=None,
    nodename=None,
    version=None,
    release=None,
    firmware=None,
) -> dict:

    "create a new empty manifest dict"
    if family is None:
        family = "micropython"  # family
    if machine is None:
        machine = family  # family

    if port is None:
        port = "common"  # family
    if platform is None:
        platform = port  # family

    if version is None:
        version = "0.0.0"

    if nodename is None:
        nodename = sysname
    if release is None:
        release = version
    if firmware is None:
        firmware = "{}-{}-{}".format(family, port, flat_version(version))

    mod_manifest = {
        "firmware": {
            "family": family,
            "port": port,
            "platform": platform,
            "machine": machine,
            "firmware": firmware,
            "nodename": nodename,
            "version": version,
            "release": release,
            "sysname": sysname,
        },
        "stubber": {"version": VERSION},
        "modules": [],
    }
    return mod_manifest


def make_manifest(folder: Path, family: str, port: str, version: str) -> bool:
    """Create a `module.json` manifest listing all files/stubs in this folder and subfolders."""
    mod_manifest = manifest(family=family, port=port, sysname=family, version=version)

    try:
        # list all *.py files, not strictly modules but decent enough for documentation
        for file in folder.glob("**/*.py"):
            mod_manifest["modules"].append({"file": str(file.relative_to(folder)), "module": file.stem})
        # write the the module manifest
        with open(os.path.join(folder, "modules.json"), "w") as outfile:
            json.dump(mod_manifest, outfile, indent=4, sort_keys=True)
        return True
    except OSError:
        return False


def generate_all_stubs():
    "just create typeshed stubs"
    # now generate typeshed files for all scripts
    print("Generate type hint files (pyi) in folder: {}".format(STUB_FOLDER))
    generate_pyi_files(Path(STUB_FOLDER))


def read_exclusion_file(path: Path = None) -> List[str]:
    """Read a .exclusion file to determine which files should not be automatically re-generated
    in .GitIgnore format

    """
    if path is None:
        path = Path(".")
    try:
        with open(path.joinpath(".exclusions")) as f:
            content = f.readlines()
            return [line.rstrip() for line in content if line[0] != "#" and len(line.strip()) != 0]
    except OSError:
        return []
    # exclusions = read_exclusion_file()


def should_ignore(file: str, exclusions: List[str]) -> bool:
    """Check if  a file matches a line in the exclusion list."""
    for excl in exclusions:
        if fnmatch(file, excl):
            return True
    return False
    # for file in Path(".").glob("**/*.py*"):
    #     if should_ignore(str(file), exclusions):
    #         print(file)



# -*- coding: utf-8 -*-

import json

import pytest
from micropy.exceptions import StubError
from micropy.stubs import source


def test_get_source(shared_datadir, test_urls, test_repo):
    """should return correct subclass"""
    test_path = shared_datadir / "esp8266_test_stub"
    local_stub = source.get_source(test_path)
    assert isinstance(local_stub, source.LocalStubSource)
    remote_stub = source.get_source("esp8266-test-stub")
    assert isinstance(remote_stub, source.RemoteStubSource)
    stub_source = source.get_source(test_urls["valid"])
    print(str(stub_source))
    assert str(stub_source) == f"<RemoteStubSource@{stub_source.location}>"


def test_source_ready(shared_datadir, test_urls, tmp_path, mocker, test_archive, test_repo):
    """should prepare and resolve stub"""
    # Test LocalStub ready
    test_path = shared_datadir / "esp8266_test_stub"
    local_stub = source.get_source(test_path)
    expected_path = local_stub.location.resolve()
    with local_stub.ready() as source_path:
        assert source_path == expected_path

    # Setup RemoteStub
    test_parent = tmp_path / "tmpdir"
    test_parent.mkdir()
    expected_path = (test_parent / "archive_test_stub").resolve()
    mocker.patch.object(source.tempfile, "mkdtemp", return_value=test_parent)
    mocker.patch.object(source.utils, "stream_download", return_value=test_archive)
    # Test Remote Stub
    remote_stub = source.get_source(test_urls["download"])
    with remote_stub.ready() as source_path:
        print(list(source_path.parent.iterdir()))
        assert (source_path / "info.json").exists()
        assert len(list(source_path.iterdir())) == 3


def test_repo_from_json(shared_datadir, mocker):
    mocker.patch.object(source.utils, "ensure_valid_url", return_value="https://testsource.com")
    test_sources = shared_datadir / "test_sources.json"
    test_repo = json.loads((shared_datadir / "test_repo.json").read_text())
    mock_get = mocker.patch.object(source.requests, "get")
    mock_get.return_value.json.return_value = test_repo
    content = test_sources.read_text()
    repos = list(source.StubRepo.from_json(content))
    assert repos[0].name == "Test Repo"
    assert len(repos) == 1


def test_repo_resolve_pkg(mocker, test_urls):
    url = test_urls["valid"]
    mocker.patch.object(source.utils, "ensure_valid_url", return_value=url)
    mocker.patch.object(source.utils, "is_downloadable", return_value=False)
    source.StubRepo("TestRepo", url, "packages")
    with pytest.raises(StubError):
        source.StubRepo.resolve_package("not-valid")


def test_repo_search(mocker, test_urls, test_repo):
    url = test_urls["valid"]
    mocker.patch.object(source.utils, "ensure_valid_url", return_value=url)
    results = test_repo.search("esp32-micropython")
    assert len(results) == 1
    assert "esp32-micropython-1.11.0" in results
    results = test_repo.search("esp32")
    assert len(results) == 2
    assert sorted(results) == sorted(["esp32-micropython-1.11.0", "esp32_LoBo-esp32_LoBo-3.2.24"])



from __future__ import annotations

import abc
from pathlib import Path
from typing import Any, AnyStr

from typing_extensions import NewType, Protocol

HostPath = NewType("HostPath", str)
DevicePath = NewType("DevicePath", str)


class StartHandler(Protocol):
    def __call__(self, *, name: str = None, size: int | None = None) -> Any:
        ...


class UpdateHandler(Protocol):
    def __call__(self, *, size: int | None = None) -> Any:
        ...


class EndHandler(Protocol):
    def __call__(self) -> Any:
        ...


class MessageHandler(Protocol):
    def __call__(self, data: AnyStr) -> Any:
        ...


class StreamConsumer(Protocol):
    @property
    @abc.abstractmethod
    def on_start(self) -> StartHandler:
        ...

    @property
    @abc.abstractmethod
    def on_update(self) -> UpdateHandler:
        ...

    @property
    @abc.abstractmethod
    def on_end(self) -> EndHandler:
        ...


class MessageConsumer(Protocol):
    @property
    @abc.abstractmethod
    def on_message(self) -> MessageHandler:
        ...


class PyDeviceConsumer(MessageConsumer, StreamConsumer, Protocol):
    ...


class MetaPyDeviceBackend(abc.ABC):
    location: str

    @abc.abstractmethod
    def establish(self, target: str) -> "MetaPyDeviceBackend":
        ...

    @abc.abstractmethod
    def connect(self) -> None:
        ...

    @abc.abstractmethod
    def disconnect(self) -> None:
        ...

    @abc.abstractmethod
    def reset(self) -> None:
        ...

    @abc.abstractmethod
    def resolve_path(self, target_path: DevicePath | str | Path) -> DevicePath:
        ...

    @property
    @abc.abstractmethod
    def connected(self) -> bool:
        ...

    @abc.abstractmethod
    def push_file(
        self,
        source_path: HostPath,
        target_path: DevicePath,
        *,
        consumer: PyDeviceConsumer | None,
        **kwargs,
    ) -> None:
        ...

    @abc.abstractmethod
    def pull_file(
        self,
        source_path: DevicePath,
        target_path: HostPath,
        *,
        consumer: PyDeviceConsumer | None,
        **kwargs,
    ) -> None:
        ...

    @abc.abstractmethod
    def list_dir(self, path: DevicePath) -> list[DevicePath]:
        ...

    @abc.abstractmethod
    def copy_dir(
        self,
        source_path: DevicePath,
        target_path: HostPath,
        *,
        consumer: PyDeviceConsumer | None,
        **kwargs,
    ):
        ...

    @abc.abstractmethod
    def eval(self, command: str, *, consumer: MessageConsumer | None = None):
        ...

    @abc.abstractmethod
    def eval_script(
        self,
        contents: AnyStr,
        target_path: DevicePath | None = None,
        *,
        consumer: PyDeviceConsumer | None = None,
    ):
        ...


class MetaPyDevice(abc.ABC):
    pydevice: MetaPyDeviceBackend
    stream_consumer: StreamConsumer | None
    message_consumer: MessageConsumer | None

    @abc.abstractmethod
    def connect(self) -> None:
        ...

    @abc.abstractmethod
    def disconnect(self) -> None:
        ...

    @abc.abstractmethod
    def copy_to(self, source_path: HostPath, target_path: DevicePath) -> None:
        ...

    @abc.abstractmethod
    def copy_from(self, source_path: DevicePath, target_path: HostPath) -> None:
        ...

    @abc.abstractmethod
    def run_script(self, content: AnyStr, target_path: DevicePath | None = None):
        ...



# -*- coding: utf-8 -*-

import subprocess

from micropy.project import checks


def test_vscode_ext_min_version(mock_checks, mocker):
    """Test VSCode Extension Template Checks"""
    assert checks.vscode_ext_min_version("ms-python.python")
    assert not checks.vscode_ext_min_version("ms-python.python", min_version="2019.9.34911")
    mock_subproc = mocker.patch.object(subprocess, "run")
    mock_subproc.side_effect = [Exception]
    assert checks.vscode_ext_min_version("ms-python.python")



"""
simple Git module, where needed via powershell
"""
from typing import Union
import subprocess
import os
from typing import Union, List


def _run_git(cmd: List[str], repo: str = None, expect_stderr=False):
    "run a external (git) command in the repo's folder and deal with some of the errors"
    try:
        if repo:
            repo = repo.replace("\\", "/")
            result = subprocess.run(cmd, capture_output=True, check=True, cwd=os.path.abspath(repo))
        else:
            result = subprocess.run(cmd, capture_output=True, check=True)
        if result.stderr != b"":
            if not expect_stderr:
                raise Exception(result.stderr.decode("utf-8"))
            print(result.stderr.decode("utf-8"))

    except subprocess.CalledProcessError as err:
        raise Exception(err)

    if result.returncode < 0:
        raise Exception(result.stderr.decode("utf-8"))
    return result


def get_tag(repo: str = None) -> Union[str, None]:
    """
    get the most recent git version tag of a local repo
    repo should be in the form of : repo = "./micropython"

    returns the tag or None
    """
    if not repo:
        repo = "."
    repo = repo.replace("\\", "/")
    cmd = ["git", "describe"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return None
    tag: str = result.stdout.decode("utf-8")
    tag = tag.replace("\r", "").replace("\n", "")
    return tag


def checkout_tag(tag: str, repo: str = None) -> bool:
    """
    get the most recent git version tag of a local repo"
    repo should be in the form of : repo = "../micropython/.git"

    returns the tag or None
    """
    cmd = ["git", "checkout", "tags/" + tag, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    print(result.stderr.decode("utf-8"))
    return True


def switch_tag(tag: str, repo: str = None) -> bool:
    """
    get the most recent git version tag of a local repo"
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns the tag or None
    """
    cmd = ["git", "switch", "--detach", tag, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    print(result.stderr.decode("utf-8"))
    return True


def switch_branch(branch: str, repo: str = None) -> bool:
    """
    get the most recent git version tag of a local repo"
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns the tag or None
    """
    cmd = ["git", "switch", branch, "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        return False
    # actually a good result
    print(result.stderr.decode("utf-8"))
    return True


def fetch(repo: str) -> bool:
    """
    fetches a repo
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns True on success
    """
    if not repo:
        raise NotADirectoryError
    repo = repo.replace("\\", "/")
    cmd = ["git", "fetch origin"]
    result = _run_git(cmd, repo=repo)
    if not result:
        return False
    return result.returncode == 0


def pull(repo: str, branch="master") -> bool:
    """
    pull a repo origin into master
    repo should be in the form of : path/.git
    repo = '../micropython/.git'
    returns True on success
    """
    if not repo:
        raise NotADirectoryError
    repo = repo.replace("\\", "/")
    # first checkout HEAD
    cmd = ["git", "checkout", "master", "--quiet", "--force"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        print("error during git checkout master", result)
        return False

    cmd = ["git", "pull", "origin", branch, "--quiet"]
    result = _run_git(cmd, repo=repo, expect_stderr=True)
    if not result:
        print("error durign pull", result)
        return False
    return result.returncode == 0



"""
Module: 'machine' on esp8266 v1.9.4
"""
# MCU: (sysname='esp8266', nodename='esp8266', release='2.2.0-dev(9422289)', version='v1.9.4-8-ga9a3caad0 on 2018-05-11', machine='ESP module with ESP8266')
# Stubber: 1.1.2


class ADC:
    ""

    def read():
        pass


DEEPSLEEP = 4
DEEPSLEEP_RESET = 5
HARD_RESET = 6


class I2C:
    ""



from __future__ import annotations

import stat
import sys
from pathlib import PurePosixPath
from typing import Type
from unittest.mock import ANY, MagicMock

import pytest
from micropy.exceptions import PyDeviceError
from micropy.pyd import backend_rshell, backend_upydevice, consumers
from micropy.pyd.abc import MetaPyDeviceBackend, PyDeviceConsumer
from micropy.pyd.pydevice import PyDevice
from pytest_mock import MockFixture
from typing_extensions import Literal


@pytest.fixture
def mock_upy(mocker: MockFixture):
    mock_upy = mocker.patch.object(backend_upydevice, "upydevice", autospec=True)
    return mock_upy


@pytest.fixture
def mock_upy_uos(mocker: MockFixture):
    mock_uos = mocker.patch.object(backend_upydevice, "UOS", autospec=True)
    return mock_uos


@pytest.fixture
def mock_rsh(mocker: MockFixture):
    mock_rsh = mocker.patch.object(backend_rshell, "rsh", autospec=True)
    return mock_rsh


class MockAdapter:
    backend: Literal["upy", "rsh"]
    mock: MagicMock
    mock_uos: MagicMock

    def __init__(self, backend: Literal["upy", "rsh"], mock: MagicMock, mock_uos=None):
        self.backend = backend
        self.mock = mock
        if mock_uos:
            self.mock_uos = mock_uos

    @property
    def is_rsh(self) -> bool:
        return self.backend == "rsh"

    @property
    def is_upy(self) -> bool:
        return self.backend == "upy"

    @property
    def connect(self) -> MagicMock:
        return self.mock.Device.return_value.connect if self.is_upy else self.mock.connect

    @property
    def device(self):
        return self.mock.Device.return_value if self.is_upy else self.mock


MOCK_PORT = "/dev/port"

IS_WIN_PY310 = sys.version_info >= (3, 10) and sys.platform.startswith("win")


@pytest.fixture(params=[True, False])
def with_consumer(request: pytest.FixtureRequest, mocker: MockFixture):
    if request.param is True:
        consumer_mock = mocker.MagicMock(PyDeviceConsumer)
        return dict(consumer=consumer_mock)
    return dict()


class TestPyDeviceBackend:
    backend: Literal["upy", "rsh"]
    pyd_cls: Type[MetaPyDeviceBackend]

    @pytest.fixture(
        params=[
            "upy",
            pytest.param(
                "rsh",
                marks=pytest.mark.skipif(
                    IS_WIN_PY310,
                    reason="skipping due to rshell/pyreadline broken for >=py310 on windows.",
                ),
            ),
        ]
    )
    def pymock_setup(self, request: pytest.FixtureRequest) -> MockAdapter:
        self.backend = request.param
        self.pyd_cls = (
            backend_upydevice.UPyDeviceBackend
            if self.backend == "upy"
            else backend_rshell.RShellPyDeviceBackend
        )

    @pytest.fixture
    def pymock(self, pymock_setup, request: pytest.FixtureRequest, mock_upy_uos):
        mod_mock = request.getfixturevalue(f"mock_{self.backend}")
        m = MockAdapter(self.backend, mod_mock, mock_upy_uos)
        yield m
        m.mock.reset_mock()

    def test_init(self, pymock):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        if self.backend == "upy":
            m.mock.Device.assert_called_once_with(MOCK_PORT, init=True, autodetect=True)
        else:
            assert m.mock.ASCII_XFER is False
            assert m.mock.QUIET is True
        assert pyd.location == MOCK_PORT

    def test_init__connect_fail(self, pymock):
        m = pymock
        m.connect.side_effect = [SystemExit, SystemExit]
        with pytest.raises(PyDeviceError):
            self.pyd_cls().establish(MOCK_PORT).connect()

    def test_connected__default(self, pymock):
        m = pymock
        pyd = self.pyd_cls()
        assert not pyd.connected
        if m.is_upy:
            with pytest.raises(PyDeviceError):
                pyd._ensure_connected()

    def test_connected(self, pymock):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.connect()
        assert pyd.connected

    def test_disconnect(self, pymock):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.disconnect()
        if m.is_upy:
            m.mock.Device.return_value.disconnect.assert_called_once()

    def test_reset(self, pymock, mocker: MockFixture):
        mocker.patch("time.sleep")
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.reset()
        if m.is_upy:
            m.device.reset.assert_called_once()
            m.device.connect.assert_called_once()

    def test_eval(self, pymock, mocker: MockFixture, with_consumer):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.connect()
        pyd._pydevice.exec_raw = mocker.Mock(return_value=["", ""])
        pyd.eval("abc", **with_consumer)
        has_cons = "consumer" in with_consumer
        if m.is_upy:
            if has_cons:
                m.device.cmd.assert_called_once_with("abc", follow=True, pipe=mocker.ANY)
            else:
                m.device.cmd.assert_called_once_with("abc")
        else:
            if has_cons:
                pyd._pydevice.exec_raw.assert_called_once_with("abc", data_consumer=mocker.ANY)
            else:
                pyd._pydevice.exec_raw.assert_called_once_with("abc", data_consumer=None)

    def test_eval_script(self, pymock, mocker: MockFixture, with_consumer):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.connect()
        pyd._pydevice.exec_raw = mocker.Mock(return_value=[b"", b""])
        pyd.eval_script(b"import something", "somefile.py", **with_consumer)
        if m.is_upy:
            if "consumer" in with_consumer:
                with_consumer["consumer"].on_start.assert_called_once()
            m.device.cmd.assert_any_call("import ubinascii")
            mock_random = mocker.patch("random.sample", return_value="abc.py")
            pyd = self.pyd_cls().establish(MOCK_PORT)
            pyd.connect()
            pyd.eval_script(b"import something", None, **with_consumer)
            mock_random.assert_called_once()

    @property
    def read_file_effects(self):
        cmd_effects = [
            None,  # import ubin
            None,  # open file
            8,  # content size
            0,  # seek start
            0,  # pos
            b"Hi there",  # read,
            8,  # pos
            None,  # close
        ]
        return cmd_effects

    def test_read_file(self, pymock):
        m = pymock
        if m.is_rsh:
            return
        pyd = self.pyd_cls().establish(MOCK_PORT)
        m.device.cmd.side_effect = self.read_file_effects
        res = pyd.read_file("/some/path")
        assert res == "Hi there"

    def test_pull_file(self, pymock, tmp_path):
        m = pymock
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.connect()
        if m.is_upy:
            m.device.cmd.side_effect = self.read_file_effects
            pyd.pull_file("/some/path", (tmp_path / "out.txt"))
            assert (tmp_path / "out.txt").read_text() == "Hi there"
        else:
            m.mock.find_serial_device_by_port.return_value.name_path = "/"
            pyd.pull_file("/some/path", (tmp_path / "out.txt"))
            m.mock.cp.assert_called_once_with("/some/path", str(tmp_path / "out.txt"))

    def test_iter_files(self, pymock):
        m = pymock
        if m.is_rsh:
            return
        pyd = self.pyd_cls().establish(MOCK_PORT)
        pyd.connect()
        m.device.cmd.side_effect = [
            None,
            None,
            None,
            [("name", "stat", "", "")],
            None,
            [("name", stat.S_IFDIR, "", "")],
            None,
            [("underName", "", "", "")],
        ]
        assert list(pyd.iter_files("/some/path")) == []
        m.device.cmd.assert_any_call("import uos", silent=True)
        m.device.cmd.assert_any_call("list(uos.ilistdir('/some/path'))", silent=True, rtn_resp=True)
        assert list(pyd.iter_files("/some/path")) == [PurePosixPath("/some/path/name")]
        assert list(pyd.iter_files("/some/path")) == [PurePosixPath("/some/path/name/underName")]


class TestPyDevice:
    @pytest.fixture
    def mock_backend(self, mocker: MockFixture):
        mock = mocker.MagicMock(MetaPyDeviceBackend)
        mock.return_value.establish.return_value = mock.return_value
        return mock

    @pytest.fixture(
        params=[
            ["dir", "/some/dir"],
            ["file", "/some/file.txt"],
            ["dir", r"c:\\some\\dos\\dir"],
            ["file", r"c:\\some\\dos\\file.txt"],
        ]
    )
    def path_type(self, request: pytest.FixtureRequest):
        return request.param

    @pytest.mark.parametrize(
        "pyd_kwargs",
        [
            dict(),
            dict(auto_connect=False),
            dict(delegate_cls=lambda *x: x, stream_consumer="stream", message_consumer="message"),
        ],
    )
    def test_init(self, mock_backend, pyd_kwargs):
        pyd = PyDevice(MOCK_PORT, backend=mock_backend, **pyd_kwargs)
        mock_backend.assert_called_once()
        mock_backend.return_value.establish.assert_called_once_with(MOCK_PORT)
        if pyd_kwargs.get("auto_connect", True):
            mock_backend.return_value.connect.assert_called_once()
        if "delegate_cls" in pyd_kwargs:
            assert pyd.consumer == (
                "stream",
                "message",
            )

    def test_connect(self, mock_backend):
        pyd = PyDevice(MOCK_PORT, backend=mock_backend, auto_connect=False)
        pyd.connect()
        mock_backend.return_value.connect.assert_called_once()

    def test_disconnect(self, mock_backend):
        pyd = PyDevice(MOCK_PORT, backend=mock_backend)
        pyd.disconnect()
        mock_backend.return_value.disconnect.assert_called_once()

    def test_copy_from(self, mock_backend, path_type, mocker):
        pyd = PyDevice(MOCK_PORT, backend=mock_backend)
        ptype, p = path_type
        pyd.copy_from(p, "/host/path")
        if ptype == "dir":
            mock_backend.return_value.copy_dir.assert_called_once_with(
                p, "/host/path", consumer=mocker.ANY
            )
        else:
            mock_backend.return_value.pull_file.assert_called_once_with(
                p, "/host/path", consumer=mocker.ANY
            )

    def test_copy_to(self, mock_backend, path_type, mocker):
        pyd = PyDevice(MOCK_PORT, backend=mock_backend)
        ptype, p = path_type
        if ptype == "dir":
            with pytest.raises(RuntimeError):
                pyd.copy_to("/host/path", p)
        else:
            pyd.copy_to("/host/path/f.txt", p)
            mock_backend.return_value.push_file.assert_called_once_with(
                "/host/path/f.txt", p, consumer=mocker.ANY
            )


class TestConsumers:
    @pytest.mark.parametrize(
        "on_desc,expected_tqdm",
        [
            [None, dict(total=5, unit="B", unit_scale=True, unit_divisor=1024, bar_format=ANY)],
            [
                lambda n, cfg: ("other", dict(override=True)),
                dict(
                    total=5,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    bar_format=ANY,
                    override=True,
                ),
            ],
        ],
    )
    def test_progress_consumer(self, mocker: MockFixture, on_desc, expected_tqdm):
        tqdm_mock = mocker.patch.object(consumers, "tqdm")
        cons = consumers.ProgressStreamConsumer(on_desc)
        cons.on_start(name="abc", size=5)
        tqdm_mock.assert_called_once_with(**expected_tqdm)
        cons.on_update(size=1)
        tqdm_mock.return_value.update.assert_called_once_with(1)
        cons.on_end()
        tqdm_mock.return_value.close.assert_called_once()

    def test_delegate(self):
        delegate = consumers.ConsumerDelegate()
        assert delegate.on_message("") is None
        delegate = consumers.ConsumerDelegate(consumers.MessageHandlers(on_message=lambda m: m))
        assert delegate.on_message("a") is "a"

    @pytest.mark.skipif(
        IS_WIN_PY310, reason="skipping due to rshell/pyreadline broken for >=py310 on windows."
    )
    def test_rsh_consumer(self):
        calls = []
        message_cons = consumers.MessageHandlers(on_message=lambda m: calls.append(m))
        rsh_cons = backend_rshell.RShellConsumer(message_cons.on_message)
        chars = iter(list("a line.\nnext line.\n"))
        for i in chars:
            rsh_cons.on_message(i.encode())
        assert len(calls) == 2
        assert calls == ["a line.", "next line."]



# -*- coding: utf-8 -*-

"""
micropy.data
~~~~~~~~~~~~~~

This module is merely to provide an easy method of locating
data files used by MicropyCli
"""

from pathlib import Path

__all__ = ["ROOT", "SCHEMAS", "REPO_SOURCES", "FILES", "STUB_DIR", "LOG_FILE", "STUBBER"]

# Paths
MOD_PATH = Path(__file__).parent
PATH = MOD_PATH.absolute()
ROOT = MOD_PATH.parent.absolute()

# Stub Schemas
SCHEMAS = PATH / "schemas"

# Default Stub Sources
REPO_SOURCES = PATH / "sources.json"

# Application Data
FILES = Path.home() / ".micropy"
STUB_DIR = FILES / "stubs"
LOG_FILE = FILES / "micropy.log"

# Libraries
LIB = ROOT / "lib"
STUBBER = LIB / "stubber"



# -*- coding: utf-8 -*-


from shutil import copytree, rmtree

import micropy.exceptions as exc
import pytest
from micropy import data, main
from micropy.stubs import stubs


def test_setup(mock_micropy_path):
    """Tests MicroPy Initial Setup"""
    expect_mp_dir = mock_micropy_path
    expect_stubs_dir = mock_micropy_path / "stubs"
    mp = main.MicroPy()
    assert expect_mp_dir.exists()
    assert expect_stubs_dir.exists()
    # Test after inital setup
    mp_ = main.MicroPy()
    assert len(mp_.stubs) == len(mp.stubs)


def test_add_stub(mock_micropy, shared_datadir):
    """Test Adding Valid Stub"""
    fware_path = shared_datadir / "fware_test_stub"
    stub_path = shared_datadir / "esp8266_test_stub"
    stubs = mock_micropy.stubs
    fware_stub = stubs.add(fware_path, data.STUB_DIR)
    stub = stubs.add(stub_path, data.STUB_DIR)
    assert stub in list(mock_micropy.stubs)
    assert stub.path in data.STUB_DIR.iterdir()
    assert stub.path.exists()
    assert fware_stub in list(mock_micropy.stubs._firmware)


def test_create_stub__connect_error(mock_micropy, mocker, shared_datadir, tmp_path):
    mock_pyb = mocker.patch("micropy.main.PyDevice")
    mock_pyb.side_effect = [SystemExit, exc.PyDeviceError]
    mp = mock_micropy
    assert mp.create_stubs("/dev/PORT") is None
    assert mp.create_stubs("/dev/PORT") is None


def test_create_stub(mock_micropy, mocker, shared_datadir, tmp_path):
    """should create and add stubs"""
    mock_micropy.stubs.add((shared_datadir / "fware_test_stub"))
    tmp_stub_path = tmp_path / "createtest"
    tmp_stub_path.mkdir()
    copytree(str(shared_datadir / "stubber_test_stub"), str(tmp_stub_path / "stubber_test_stub"))
    mock_tmpdir = mocker.patch("micropy.main.tempfile.TemporaryDirectory")
    mock_tmpdir.return_value.__enter__.return_value = tmp_stub_path
    mock_pyb = mocker.patch("micropy.main.PyDevice")
    mp = mock_micropy
    mocker.spy(mp.stubs, "add")
    mock_pyb.return_value.run_script.side_effect = [Exception, mocker.ANY, mocker.ANY]
    with pytest.raises(Exception):
        stub = mp.create_stubs("/dev/PORT")
    stub = mp.create_stubs("/dev/PORT")
    mp.stubs.add.assert_any_call((tmp_stub_path / "esp32-1.11.0"))
    rmtree((tmp_stub_path / "esp32-1.11.0"))
    assert isinstance(stub, stubs.DeviceStub)


def test_stub_error():
    with pytest.raises(exc.StubError):
        raise exc.StubError(None)


def test_resolve_project(mocker, mock_micropy):
    mock_proj = mocker.patch.object(main, "Project").return_value
    mock_proj.exists = False
    assert not mock_micropy.resolve_project(".").exists
    mock_proj.exists = True
    assert mock_micropy.resolve_project(".")



# make_stub_files: Thu 20 Jun 2019 at 23:08:04

from typing import Any, Dict, Optional, Sequence, Tuple, Union

Node = Any
class TokenInfo(namedtuple(str, Tuple[str, str, str, str, str])):
    def __str__(self) -> str: ...
def get_indent(l: Any) -> Tuple[int, Any]: ...
def tokenize(readline: Any) -> None: ...



# This file is executed on every boot (including wake-boot from deepsleep)
import esp
import machine
import uos as os

esp.osdebug(None)

SD = False
if SD:
    # Mount SD to /sd
    try:
        # Some boards have pulldown and/or LED on GPIO2, pullup avoids issues on TTGO 8 v1.8
        # machine.Pin(2,mode=machine.Pin.IN, pull=machine.Pin.PULL_UP)
        # os.mount(machine.SDCard(slot=1, width=4), "/sd")  # SD mode 4 bit

        # # SPI 1 bit M5Stack Core
        os.mount(machine.SDCard(slot=2, width=1, sck=18, miso=19, mosi=23, cs=4), "/sd")  # SPI 1 bit M5Stack Core
        print("SD Card mounted")
    except OSError as e:
        if e.args[0] == 16:
            print("No SD Card found")



# -*- coding: utf-8 -*-

"""Micropy Cli.

Micropy Cli is a project management/generation tool for writing Micropython
code in modern IDEs such as VSCode. Its primary goal is to automate the
process of creating a workspace complete with:

Linting compatible with Micropython,
VSCode Intellisense,
Autocompletion,
Dependency Management,
VCS Compatibility
and more.

"""

from micropy import data, lib, project, pyd, stubs, utils

from .main import MicroPy

__author__ = """Braden Mars"""
__version__ = "4.0.0-rc.2"

__all__ = ["MicroPy", "data", "lib", "project", "stubs", "utils", "pyd"]



c='utf8'
b='{}/{}'
a='_thread'
Z=IndexError
Y=NameError
X=print
W=NotImplementedError
R='__init__'
P='version'
O=True
N=KeyError
M=ImportError
L='.'
K='_'
J=open
I=len
H=''
G=False
F=AttributeError
D='/'
C=None
A=OSError
import sys,gc as E,uos as os
from utime import sleep_us as d
from ujson import dumps as Q
__version__='1.4.3'
S=2
e=2
try:from machine import resetWDT as T
except M:
	def T():0
class Stubber:
	def __init__(B,path=C,firmware_id=C):
		G=firmware_id;C=path
		try:
			if os.uname().release=='1.13.0'and os.uname().version<'v1.13-103':raise W('MicroPython 1.13.0 cannot be stubbed')
		except F:pass
		B._report=[];B.info=_info();E.collect()
		if G:B._fwid=str(G).lower()
		else:B._fwid='{family}-{port}-{ver}'.format(**B.info).lower()
		B._start_free=E.mem_free()
		if C:
			if C.endswith(D):C=C[:-1]
		else:C=get_root()
		B.path='{}/stubs/{}'.format(C,B.flat_fwid).replace('//',D)
		try:U(C+D)
		except A:pass
		B.problematic=['upip','upysh','webrepl_setup','http_client','http_client_ssl','http_server','http_server_ssl'];B.excluded=['webrepl','_webrepl','port_diag','example_sub_led.py','example_pub_button.py'];B.modules=[]
	def get_obj_attributes(J,item_instance):
		I="Couldn't get attribute '{}' from object '{}', Err: {}";A=item_instance;B=[];D=[]
		try:
			for C in dir(A):
				try:G=getattr(A,C);B.append((C,repr(G),repr(type(G)),G))
				except F as H:D.append(I.format(C,A,H))
		except F as H:D.append(I.format(C,A,H))
		B=[A for A in B if not(A[0].startswith(K)and A[0]!=R)];E.collect();return B,D
	def add_modules(A,modules):A.modules=sorted(set(A.modules)|set(modules))
	def create_all_stubs(A):
		E.collect()
		for B in A.modules:A.create_one_stub(B)
	def create_one_stub(C,module_name):
		B=module_name
		if B.startswith(K)and B!=a:return G
		if B in C.problematic:return G
		if B in C.excluded:return G
		F='{}/{}.py'.format(C.path,B.replace(L,D));E.collect();H=E.mem_free();X('Stub module: {:<20} to file: {:<55} mem:{:>5}'.format(B,F,H))
		try:C.create_module_stub(B,F)
		except A:return G
		E.collect();return O
	def create_module_stub(F,module_name,file_name=C):
		G=file_name;B=module_name
		if B.startswith(K)and B!=a:return
		if B in F.problematic:return
		if G is C:G=F.path+D+B.replace(L,K)+'.py'
		if D in B:B=B.replace(D,L)
		I=C
		try:I=__import__(B,C,C,'*')
		except M:return
		U(G)
		with J(G,'w')as O:P='"""\nModule: \'{0}\' on {1}\n"""\n# MCU: {2}\n# Stubber: {3}\n'.format(B,F._fwid,F.info,__version__);O.write(P);O.write('from typing import Any\n\n');F.write_object_stub(O,I,B,H)
		F._report.append({'module':B,'file':G})
		if not B in['os','sys','logging','gc']:
			try:del I
			except (A,N):pass
			try:del sys.modules[B]
			except N:pass
		E.collect()
	def write_object_stub(L,fp,object_expr,obj_name,indent,in_class=0):
		b='tuple';a='list';Z='dict';X='{0}{1} = {2} # type: {3}\n';W='bound_method';V='Any';P=in_class;O=object_expr;J=fp;B=indent;E.collect()
		if O in L.problematic:return
		Q,c=L.get_obj_attributes(O)
		for (C,K,F,S) in Q:
			if C in['classmethod','staticmethod']:continue
			T();d(1)
			if F=="<class 'type'>"and I(B)<=e*4:D='\n'+B+'class '+C+':\n';D+=B+"    ''\n";J.write(D);L.write_object_stub(J,S,'{0}.{1}'.format(obj_name,C),B+'    ',P+1)
			elif'method'in F or'function'in F or C==R:
				M=V;U=H
				if P>0:
					U='self, '
					if C==R:M='None'
				if W in F or W in K:D='{}@classmethod\n'.format(B);D+='{}def {}(cls, *args) -> {}:\n'.format(B,C,M)
				else:D='{}def {}({}*args) -> {}:\n'.format(B,C,U,M)
				D+=B+'    ...\n\n';J.write(D)
			elif F=="<class 'module'>":0
			elif F.startswith("<class '"):
				G=F[8:-2];D=H
				if G in['str','int','float','bool','bytearray','bytes']:D=X.format(B,C,K,G)
				elif G in[Z,a,b]:f={Z:'{}',a:'[]',b:'()'};D=X.format(B,C,f[G],G)
				else:
					if not G in['object','set','frozenset']:G=V
					D='{0}{1} : {2} ## {3} = {4}\n'.format(B,C,G,F,K)
				J.write(D)
			else:J.write("# all other, type = '{0}'\n".format(F));J.write(B+C+' # type: Any\n')
		del Q;del c
		try:del C,K,F,S
		except (A,N,Y):pass
	@property
	def flat_fwid(self):
		A=self._fwid;B=' .()/\\:$'
		for C in B:A=A.replace(C,K)
		return A
	def clean(E,path=C):
		B=path
		if B is C:B=E.path
		X('Clean/remove files in folder: {}'.format(B))
		try:G=os.listdir(B)
		except (A,F):return
		for H in G:
			D=b.format(B,H)
			try:os.remove(D)
			except A:
				try:E.clean(D);os.rmdir(D)
				except A:pass
	def report(C,filename='modules.json'):
		D=',\n';H=b.format(C.path,filename);E.collect()
		try:
			with J(H,'w')as B:
				B.write('{');B.write(Q({'firmware':C.info})[1:-1]);B.write(D);B.write(Q({'stubber':{P:__version__}})[1:-1]);B.write(D);B.write('"modules" :[\n');F=O
				for I in C._report:
					if F:F=G
					else:B.write(D)
					B.write(Q(I))
				B.write('\n]}')
			K=C._start_free-E.mem_free()
		except A:pass
def U(path):
	C=path;B=F=0
	while B!=-1:
		B=C.find(D,F)
		if B!=-1:
			if B==0:E=C[0]
			else:E=C[0:B]
			try:I=os.stat(E)
			except A as G:
				if G.args[0]==S:
					try:os.mkdir(E)
					except A as H:raise H
				else:raise G
		F=B+1
def _info():
	g='loboris';f=' on ';e='0.0.0';d='port';c='platform';b='machine';a='nodename';Y='name';T='v';S='mpy';R='unknown';Q='-';O='sysname';J='ver';G='family';E='build';B='release';U=sys.implementation.name;V=sys.platform;A={Y:U,B:e,P:e,E:H,O:R,a:R,b:R,G:U,c:V,d:V,J:H}
	try:A[B]=L.join([str(A)for A in sys.implementation.version]);A[P]=A[B];A[Y]=sys.implementation.name;A[S]=sys.implementation.mpy
	except F:pass
	if sys.platform not in('unix','win32'):
		try:
			D=os.uname();A[O]=D.sysname;A[a]=D.nodename;A[B]=D.release;A[b]=D.machine
			if f in D.version:
				K=D.version.split(f)[0]
				if A[O]=='esp8266':
					if Q in K:W=K.split(Q)[0]
					else:W=K
					A[P]=A[B]=W.lstrip(T)
				try:A[E]=K.split(Q)[1]
				except Z:pass
		except (Z,F,TypeError):pass
	try:from pycopy import const;A[G]='pycopy';del const
	except (M,N):pass
	if A[c]=='esp32_LoBo':A[G]=g;A[d]='esp32'
	elif A[O]=='ev3':
		A[G]='ev3-pybricks';A[B]='1.0.0'
		try:from pybricks.hubs import EV3Brick;A[B]='2.0.0'
		except M:pass
	if A[B]:A[J]=T+A[B].lstrip(T)
	if A[G]!=g:
		if A[B]and A[B]>='1.10.0'and A[B].endswith('.0'):A[J]=A[B][:-2]
		else:A[J]=A[B]
		if A[E]!=H and I(A[E])<4:A[J]+=Q+A[E]
	if S in A:
		h=int(A[S]);X=[C,'x86','x64','armv6','armv6m','armv7m','armv7em','armv7emsp','armv7emdp','xtensa','xtensawin'][h>>10]
		if X:A['arch']=X
	return A
def get_root():
	B='/flash'
	try:E=os.stat(B)
	except A as C:
		if C.args[0]==S:
			try:B=os.getcwd()
			except (A,F):B=L
		else:B=D
	return B
def B():sys.exit(1)
def read_path():
	A=H
	if I(sys.argv)==3:
		C=sys.argv[1].lower()
		if C in('--path','-p'):A=sys.argv[2]
		else:B()
	elif I(sys.argv)==2:B()
	return A
def V():
	try:A=bytes('abc',encoding=c);B=V.__module__;return G
	except (W,F):return O
def f():
	M=b'todo';L='.db';F='modulelist';import machine as N,btree
	try:D=J(F+L,'r+b');E=O
	except A:D=J(F+L,'w+b');E=G
	stubber=Stubber(path=read_path())
	if not E:stubber.clean()
	B=btree.open(D)
	if not E or I(list(B.keys()))==0:
		for P in J(F+'.txt'):
			C=P.strip()
			if I(C)and C[0]!='#':B[C]=M
		B.flush()
	for C in B.keys():
		if B[C]!=M:continue
		H=G
		try:H=stubber.create_one_stub(C.decode(c))
		except MemoryError:B.close();D.close();N.reset()
		if H:K='good, I guess'
		else:K=b'skipped'
		B[C]=K;B.flush()
	B.close();D.close()
if __name__=='__main__'or V():
	try:logging.basicConfig(level=logging.INFO)
	except Y:pass
	f()


# -*- coding: utf-8 -*-

"""
micropy.utils.validate
~~~~~~~~~~~~~~

This module contains utility functions for MicropyCli
that center on data validation
"""

import json
from pathlib import Path

from jsonschema import Draft7Validator


class Validator:
    """ "jsonschema wrapper for file validation.

    Returns:
        object: Validator Instance

    """

    def __init__(self, schema_path):
        schema = self._load_json(schema_path)
        self.schema = Draft7Validator(schema)

    def _load_json(self, path):
        """Loads json data from file.

        Args:
            path (str): path to file

        Returns:
            Loaded JSON data as an array or dict

        """
        file = Path(path).resolve()
        data = json.load(file.open())
        return data

    def validate(self, path):
        """Validates json file against a schema.

        Args:
            path (str): path to json file to validate

        Returns:
            jsonschema.validate

        """
        data = self._load_json(path)
        return self.schema.validate(data)



# -*- coding: utf-8 -*-

from pathlib import Path

import click
import pytest
from click.testing import CliRunner
from micropy import cli
from micropy.exceptions import RequirementException


@pytest.fixture
def runner():
    runner = CliRunner()
    return runner


@pytest.fixture
def mock_mpy(mocker):
    mock_mp = mocker.patch("micropy.main.MicroPy")
    mocker.patch.object(click.Context, "find_object").return_value = mock_mp
    return mock_mp


def test_cli_micropy(runner, mocker, mock_cwd):
    """should execute"""
    if (mock_cwd / ".micropy").exists():
        (mock_cwd / ".micropy").unlink()
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    expected = "CLI Application for creating/managing" " Micropython Projects."
    assert expected in result.output
    # should alert about update
    mocker.patch.object(cli.utils, "is_update_available", return_value="1.0.0")
    log_spy = mocker.spy(cli, "Log")
    runner.invoke(cli.cli, ["stubs"])
    log_spy.get_logger.assert_called_once_with("MicroPy")


def test_stub_list(mock_mpy, mocker, runner):
    """should list stubs"""
    mock_mpy.log.title = print
    mock_mpy.log.info = print
    m_data = [("FakeFirmware", ["dev1", "dev2"])]
    mock_project = mocker.patch.object(cli, "Project")
    mock_mpy.resolve_project.return_value = mock_project.return_value
    mock_mpy.project.name = "Test Project"
    mock_mpy.project.stubs = m_data
    mock_mpy.stubs.iter_by_firmware.return_value = m_data
    result = runner.invoke(cli.list)
    assert result.exit_code == 0
    for s in ["dev1", "dev2"]:
        assert s in result.output


def test_stub_create(runner, mock_mpy):
    """should call create_stubs"""
    result = runner.invoke(cli.create, ["/dev/PORT"], obj=mock_mpy)
    mock_mpy.create_stubs.assert_called_once_with("/dev/PORT", verbose=False)
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "cliargs,expargs",
    [
        (
            "TestProject -t vscode",
            {
                "project": {"path": "TestProject", "name": None},
                "template": {"templates": ("vscode",)},
            },
        ),
        (
            "-t vscode",
            {
                "project": {"path": Path.cwd(), "name": "TestProject"},
                "template": {"templates": ("vscode",)},
            },
        ),
    ],
)
def test_cli_init(mocker, mock_mpy, shared_datadir, mock_prompt, runner, cliargs, expargs):
    """should create project"""
    # Mock Project
    mock_project = mocker.patch.object(cli, "Project")
    mock_modules = mocker.patch.object(cli, "modules")
    # Mock Text Prompt
    ptext_mock = mocker.patch.object(cli.prompt, "text").return_value
    ptext_mock.ask.return_value = "TestProject"
    # Test with no stubs (should fail)
    result = runner.invoke(cli.init, cliargs.split())
    assert result.exit_code == 1
    # Test with Stubs
    mock_mpy.stubs = ["stub"]
    result = runner.invoke(cli.init, cliargs.split())
    # Assert Project
    exp_project = expargs.pop("project", {})
    exp_path = exp_project.pop("path", mocker.ANY)
    mock_project.assert_called_once_with(exp_path, **exp_project)
    # mock_add = mock_project.return_value.add
    # Assert Templates
    expect_calls = [
        mocker.call(mock_modules.StubsModule, mock_mpy.stubs, stubs=["stub"]),
        mocker.call(mock_modules.PackagesModule, "requirements.txt"),
        mocker.call(mock_modules.DevPackagesModule, "dev-requirements.txt"),
        mocker.call(mock_modules.TemplatesModule, templates=("vscode",), run_checks=mocker.ANY),
    ]
    mock_project.return_value.add.assert_has_calls(expect_calls, any_order=True)
    # # Assert Exit Code
    mock_project.return_value.create.assert_called_once()
    assert result.exit_code == 0


def test_cli_stubs_add(mocker, mock_mpy, shared_datadir, runner, tmp_path, mock_checks):
    """should add stub"""
    mock_proj = mocker.patch.object(cli, "Project").return_value
    mock_proj.exists.return_value = True
    mock_mpy.project = mock_proj
    mock_mpy.stubs.add.side_effect = [cli.exc.StubError, cli.exc.StubNotFound, mocker.MagicMock()]
    mock_mpy.log.error = print

    mocker.spy(cli.sys, "exit")

    result = runner.invoke(cli.add, ["invalid-stub"], obj=mock_mpy)
    assert cli.sys.exit.called_with(1)
    assert "is not a valid stub" in result.output
    assert result.exit_code == 1

    result = runner.invoke(cli.add, ["not-real-stub"])
    assert cli.sys.exit.called_with(1)
    assert "could not be found" in result.output
    assert result.exit_code == 1

    result = runner.invoke(cli.add, ["real-stub"])
    assert mock_proj.add_stub.call_count == 1
    assert result.exit_code == 0


def test_cli_stubs_search(mock_mpy, runner):
    """should search stubs"""
    mock_mpy.stubs.search_remote.return_value = [
        (
            "esp8266-micropython-1.11.0",
            True,
        ),
        (
            "esp8266-micropython-1.10.0",
            False,
        ),
    ]
    result = runner.invoke(cli.search, ["esp8266"])
    assert result.exit_code == 0


class TestInstall:
    @pytest.fixture
    def mock_proj(self, mocker, mock_mpy):
        mock_project = mocker.patch.object(cli, "Project")
        mock_proj = mock_project.return_value
        mock_mpy.project = mock_proj
        return mock_proj

    def test_normal(self, runner, mock_proj):
        result = runner.invoke(cli.install, ["package", "--dev"])
        assert result.exit_code == 0
        mock_proj.add_package.assert_called_once_with("package", dev=True)

    def test_from_requirements(self, runner, mock_proj):
        result = runner.invoke(cli.install, "")
        assert result.exit_code == 0
        mock_proj.add_from_file.side_effect = [FileNotFoundError]
        result = runner.invoke(cli.install, "")
        assert result.exit_code == 1
        assert "Aborted!" in result.output

    def test_no_project_found(self, runner):
        result = runner.invoke(cli.install, ["package"])
        assert result.exit_code == 1

    def test_bad_package_name(self, runner, mock_proj):
        mock_proj.add_package.side_effect = [RequirementException]
        result = runner.invoke(cli.install, ["badpackage"])
        assert result.exit_code == 1
        assert "Aborted!" in result.output

    def test_from_path(self, runner, mock_proj, tmp_path):
        tmp_package = tmp_path / "mycustompackage"
        tmp_package.mkdir()
        result = runner.invoke(cli.install, ["--path", str(tmp_package)])
        assert result.exit_code == 0



# -*- coding: utf-8 -*-

import json
from pathlib import Path

import pylint.lint
import pytest
from micropy.project.template import Template, TemplateProvider


@pytest.fixture
def stub_context(mock_mp_stubs):
    stubs = list(mock_mp_stubs.stubs)[:3]
    stub_paths = [stub.stubs for stub in stubs]
    frozen_paths = [stub.frozen for stub in stubs]
    fware_paths = [stub.firmware.frozen for stub in stubs]
    ctx_paths = [*stub_paths, *frozen_paths, *fware_paths]
    return (stubs, (stub_paths, frozen_paths, fware_paths), ctx_paths)


def test_vscode_template(stub_context, shared_datadir, tmp_path, mock_checks):
    stubs, paths, ctx_paths = stub_context
    prov = TemplateProvider(["vscode"])
    ctx_datadir = tmp_path / "ctx_cata"
    ctx_datadir.mkdir(exist_ok=True)
    # Add test local path
    ctx_local = ctx_datadir / "src" / "lib" / "somelib"
    ctx_local.mkdir(parents=True)
    ctx_absolute = Path("/fakedir/notinprojectdir/somelib")
    ctx_local_paths = [ctx_local, ctx_absolute]
    prov.render_to(
        "vscode",
        tmp_path,
        stubs=stubs,
        paths=ctx_paths,
        datadir=ctx_datadir,
        local_paths=ctx_local_paths,
    )
    expected_path = tmp_path / ".vscode" / "settings.json"
    out_content = expected_path.read_text()
    print(out_content)
    # Get rid of comments
    with expected_path.open() as f:
        lines = [l.strip() for l in f.readlines() if l]
        valid = [l for l in lines if "//" not in l[:2]]
    # Valid JSON?
    expect_paths = [str(p.relative_to(tmp_path)) for p in ctx_paths]
    expect_paths.append(str(ctx_local.relative_to(tmp_path)))  # add local path (should be relative)
    # local path outside of project dir (must be absolute)
    expect_paths.append(str(ctx_absolute.absolute()))
    content = json.loads("\n".join(valid))
    assert sorted(expect_paths) == sorted(content["python.autoComplete.extraPaths"])
    assert expected_path.exists()
    # Test Update
    ctx_paths.append((tmp_path / "foobar" / "foo.py"))
    prov.update(
        "vscode",
        tmp_path,
        stubs=stubs,
        paths=ctx_paths,
        datadir=ctx_datadir,
        local_paths=ctx_local_paths,
    )
    content = json.loads(expected_path.read_text())
    expect_paths.append(str((tmp_path / "foobar" / "foo.py").relative_to(tmp_path)))
    assert sorted(expect_paths) == sorted(content["python.autoComplete.extraPaths"])
    # Test update with missing file
    expected_path.unlink()  # delete file
    prov.update("vscode", tmp_path, stubs=stubs, paths=ctx_paths, datadir=ctx_datadir)
    assert expected_path.exists()


def test_pylint_template(stub_context, tmp_path):
    def test_pylint_load():
        try:
            lint_args = ["--rcfile", str(expected_path.absolute())]
            pylint.lint.Run(lint_args)
        except SyntaxError:
            pytest.fail(str(SyntaxError))  # noqa
        except:  # noqa
            pass

    stubs, paths, ctx_paths = stub_context
    ctx_datadir = tmp_path / "ctx_cata"
    ctx_datadir.mkdir(exist_ok=True)
    prov = TemplateProvider(["pylint"])
    prov.render_to("pylint", tmp_path, stubs=stubs, paths=ctx_paths, datadir=ctx_datadir)
    expected_path = tmp_path / ".pylintrc"
    assert expected_path.exists()
    # Will Pylint load it?
    test_pylint_load()
    # Test Update
    new_path = tmp_path / ".micropy" / "foobar" / "foo"
    ctx_paths.append(new_path)
    prov.update("pylint", tmp_path, stubs=stubs, paths=ctx_paths, datadir=ctx_datadir)
    init_hook = expected_path.read_text().splitlines(True)[2]
    hook_imports = init_hook.split(",")
    hook_path = str(Path(".micropy/foobar/foo")).replace(
        "\\", "/"
    )  # no need to use \\ on pylint Windows
    assert f' "{hook_path}"' in hook_imports
    test_pylint_load()


def test_generic_template(mock_mp_stubs, tmp_path):
    prov = TemplateProvider(["bootstrap", "pymakr"])
    prov.render_to("boot", tmp_path)
    expected_path = tmp_path / "src" / "boot.py"
    assert expected_path.exists()
    expected_content = (prov.TEMPLATE_DIR / "src" / "boot.py").read_text()
    out_content = expected_path.read_text()
    print(out_content)
    assert expected_content.strip() == out_content.strip()
    templ = prov.get("boot")
    assert templ.update(tmp_path) is None


def test_no_context():
    class BadTemplate(Template):
        def __init__(self, template, **kwargs):
            return super().__init__(template, **kwargs)

    with pytest.raises(NotImplementedError):
        x = BadTemplate("abc")
        print(x.context)



from __future__ import annotations

import binascii
import io
import random
import stat
import string
import time
from functools import wraps
from pathlib import Path, PurePosixPath
from typing import AnyStr, Callable, Generator, TypeVar, Union

import upydevice
from boltons import iterutils
from micropy.exceptions import PyDeviceConnectionError, PyDeviceError
from typing_extensions import ParamSpec, TypeAlias
from upydevice.phantom import UOS as UPY_UOS

from .abc import DevicePath, HostPath, MessageConsumer, MetaPyDeviceBackend, PyDeviceConsumer

AnyUPyDevice: TypeAlias = Union[upydevice.SerialDevice, upydevice.WebSocketDevice]

BUFFER_SIZE = 512

T = TypeVar("T")
P = ParamSpec("P")


class UOS(UPY_UOS):
    @upydevice.upy_cmd_c_r()
    def stat(self, path):
        return self.dev_dict


def retry(fn: Callable[P, T]) -> Callable[P, T | None]:
    @wraps(fn)
    def _wrapper(self_: UPyDeviceBackend, *args: P.args, **kwargs: P.kwargs) -> T | None:
        _result: T | None = None
        retry_count = 0

        while retry_count < 4:
            try:
                _result = fn(self_, *args, **kwargs)  # type: ignore
            except Exception as e:
                retry_count += 1
                self_.BUFFER_SIZE = BUFFER_SIZE // pow(2, retry_count + 1)
                print("reducing buffer size to:", self_.BUFFER_SIZE)
                print(e)
                self_.reset()
            else:
                break
        return _result

    return _wrapper  # type: ignore


class UPyDeviceBackend(MetaPyDeviceBackend):
    BUFFER_SIZE: int = BUFFER_SIZE

    _pydevice: AnyUPyDevice
    _uos: UOS | None = None

    def _ensure_connected(self):
        if not self.connected:
            raise PyDeviceError(f"No currently connected device found!")

    def _rand_device_path(self) -> DevicePath:
        name = "".join(random.sample(string.ascii_lowercase, 6))
        return self.resolve_path(DevicePath(name))

    @property
    def uos(self) -> UOS:
        self._ensure_connected()
        if not self._uos:
            self._uos = UOS(self._pydevice)
        return self._uos

    def _pyb_root(self) -> DevicePath:
        results = self.uos.stat("/flash")
        if "Traceback" or "ENOENT" in results:
            return DevicePath("/")
        return DevicePath("/flash")

    def resolve_path(self, path: DevicePath | str | Path) -> DevicePath:
        _root = PurePosixPath(self._pyb_root())
        _path = PurePosixPath(path)
        if _path.is_absolute():
            if _root == _path or _root in list(_path.parents):
                return DevicePath(str(_path))
            _path = _path.relative_to(list(_path.parents)[-1])
        return DevicePath(str(_root / _path))

    def establish(self, target: str) -> UPyDeviceBackend:
        self.location = target
        self._pydevice = upydevice.Device(target, init=True, autodetect=True)
        return self

    def connect(self):
        try:
            self._pydevice.connect()
        except (SystemExit, Exception) as e:
            raise PyDeviceConnectionError(self.location) from e

    def disconnect(self):
        if self.connected:
            self._pydevice.disconnect()

    def reset(self):
        self._pydevice.reset()
        time.sleep(2)
        self._pydevice.connect()
        time.sleep(4)

    @property
    def connected(self) -> bool:
        return False if getattr(self, "_pydevice", None) is None else self._pydevice.connected

    def list_dir(self, path: DevicePath) -> list[DevicePath]:
        return [DevicePath(p) for p in self.uos.listdir(self.resolve_path(path))]

    def iter_files(self, path: DevicePath) -> Generator[DevicePath, None, None]:
        path = self.resolve_path(path)
        self._pydevice.cmd("import uos", silent=True)
        results = self._pydevice.cmd(f"list(uos.ilistdir('{path}'))", silent=True, rtn_resp=True)
        if not results:
            return
        for name, type_, _, _ in results:
            abs_path = PurePosixPath(path) / name
            if type_ == stat.S_IFDIR:
                yield from self.iter_files(abs_path)
            else:
                yield abs_path

    def copy_dir(self, source_path: DevicePath, target_path: HostPath, **kwargs):
        target_path = Path(str(target_path))  # type: ignore
        source_path = self.resolve_path(source_path)
        for file_path in self.iter_files(source_path):
            rel_path = PurePosixPath(file_path).relative_to(
                list(PurePosixPath(file_path).parents)[-1]
            )
            # handles os-path conversion
            file_dest = Path(target_path / rel_path)
            file_dest.parent.mkdir(parents=True, exist_ok=True)
            self.pull_file(file_path, HostPath(str(file_dest)), **kwargs)

    def push_file(self, source_path: HostPath, target_path: DevicePath, **kwargs) -> None:
        src_path = Path(str(source_path))
        src_contents = src_path.read_text()
        self.write_file(src_contents, target_path, **kwargs)

    def pull_file(self, source_path: DevicePath, target_path: HostPath, **kwargs) -> None:
        src_path = self.resolve_path(source_path)
        targ_path = Path(str(target_path))
        source_contents = self.read_file(src_path, **kwargs)
        if source_contents is None:
            # TODO: properly report failure to read/copy file.
            return None
        targ_path.write_text(source_contents)

    def _iter_hex_chunks(self, content: str):
        chunked_content = iterutils.chunked_iter(content, self.BUFFER_SIZE)
        for chunk in chunked_content:
            hex_chunk = binascii.hexlify(chunk.encode())
            yield hex_chunk

    @retry
    def write_file(
        self, contents: str, target_path: DevicePath, *, consumer: PyDeviceConsumer | None = None
    ) -> None:
        target_path = self.resolve_path(target_path)
        self._pydevice.cmd("import ubinascii")
        self._pydevice.cmd(f"f = open('{str(target_path)}', 'wb')")
        content_size = len(binascii.hexlify(contents.encode())) // 2
        if consumer:
            consumer.on_start(name=f"Writing {str(target_path)}", size=content_size)
        for chunk in self._iter_hex_chunks(contents):
            cmd = f"contents = ubinascii.unhexlify('{chunk.decode()}'); f.write(contents)"
            self._pydevice.cmd(cmd, silent=True)
            if consumer:
                consumer.on_update(size=len(chunk) // 2)
        if consumer:
            consumer.on_end()
        self._pydevice.cmd("f.close()")

    @retry
    def read_file(
        self, target_path: DevicePath, *, consumer: PyDeviceConsumer | None = None
    ) -> str:
        target_path = self.resolve_path(target_path)
        self._pydevice.cmd("import ubinascii", silent=True)
        self._pydevice.cmd(f"f = open('{str(target_path)}', 'rb')", silent=True)
        content_size = self._pydevice.cmd(f"f.seek(0,2)", rtn_resp=True, silent=True)
        self._pydevice.cmd("f.seek(0)", silent=True)
        buffer = io.BytesIO()
        if consumer:
            consumer.on_start(name=f"Reading {target_path}", size=content_size // 2)
        last_pos = 0
        while True:
            pos = self._pydevice.cmd(f"f.tell()", rtn_resp=True, silent=True)
            if consumer:
                consumer.on_update(size=(pos - last_pos) // 2)
            last_pos = pos
            if pos == content_size:
                if consumer:
                    consumer.on_end()
                break
            next_chunk = self._pydevice.cmd(
                f"f.read({self.BUFFER_SIZE})", rtn_resp=True, silent=True
            )
            buffer.write(next_chunk)
        self._pydevice.cmd("f.close()")
        value = buffer.getvalue().decode()
        return value

    def eval(self, command: str, *, consumer: MessageConsumer = None):
        if consumer:
            return self._pydevice.cmd(
                command, follow=True, pipe=lambda m, *args, **kws: consumer.on_message(m)
            )
        return self._pydevice.cmd(command)

    def eval_script(
        self,
        contents: AnyStr,
        target_path: DevicePath | None = None,
        *,
        consumer: PyDeviceConsumer = None,
    ):
        _target_path = (
            self.resolve_path(target_path) if target_path else f"{self._rand_device_path()}.py"
        )
        if isinstance(contents, bytes):
            contents = contents.decode()  # type: ignore
        self.write_file(str(contents), DevicePath(_target_path), consumer=consumer)
        self.eval(f"import {Path(_target_path).stem}", consumer=consumer)
        self.uos.remove(str(_target_path))



# -*- coding: utf-8 -*-

import abc
import inspect
from copy import deepcopy
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple, Type, TypeVar, Union

from micropy import utils
from micropy.config import Config
from micropy.logger import Log, ServiceLog

"""Project Packages Module Abstract Implementation"""

T = TypeVar("T")
ProxyItem = List[Tuple[T, str]]


class ProjectModule(metaclass=abc.ABCMeta):
    """Abstract Base Class for Project Modules."""

    PRIORITY: int = 0
    _hooks: List["HookProxy"] = []

    def __init__(self, parent: Optional["ProjectModule"] = None, log: Optional[ServiceLog] = None):
        self._parent = parent
        self.log = log

    @property
    def parent(self):
        """Component Parent."""
        return self._parent

    @parent.setter
    def parent(self, parent: Type["ProjectModule"]) -> Type["ProjectModule"]:
        """Sets component parent.

        Args:
            parent (Any): Parent to set

        """
        self._parent = parent
        return self.parent

    @abc.abstractproperty
    def config(self) -> Union[dict, Config]:
        """Config values specific to component."""

    @abc.abstractmethod
    def load(self):
        """Method to load component."""

    @abc.abstractmethod
    def create(self, *args: Any, **kwargs: Any) -> Any:
        """Method to create component."""

    @abc.abstractmethod
    def update(self):
        """Method to update component."""

    def add(self, component: Type["ProjectModule"], *args: Any, **kwargs: Any) -> Any:
        """Adds component.

        Args:
            component (Any): Component to add.

        """

    def remove(self, component: Type["ProjectModule"]) -> Any:
        """Removes component.

        Args:
            component (Any): Component to remove.

        """

    @classmethod
    def hook(cls, *args: Any, **kwargs: Any) -> Callable[..., Any]:
        """Decorator for creating a Project Hook.

        Allows decorated method to be called from parent
        container.

        Returns:
            Callable: Decorated function.

        """

        def _hook(func: T) -> Callable[..., Any]:
            name = kwargs.get("name", func.__name__)
            hook = next((i for i in cls._hooks if i._name == name), None)
            if not hook:
                hook = HookProxy(name)
                ProjectModule._hooks.append(hook)
            hook.add_method(func, **kwargs)

            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> T:
                return func(*args, **kwargs)

            return wrapper

        return _hook

    def resolve_hook(self, name: str) -> Union[Optional["HookProxy"], T]:
        """Resolves appropriate hook for attribute name.

        Args:
            name (str): Attribute name to resolve hook for.

        Returns:
            Optional[HookProxy]: Callable Proxy for ProjectHook.
            NoneType: Name could not be resolved.

        """
        _hook = None
        for hook in self._hooks:
            if hook._name == name:
                _hook = hook
                _hook.add_instance(self)
                if _hook.is_descriptor():
                    return _hook.get()
        return _hook


class HookProxy:
    """Proxy for Project Hooks.

    Allows multiple project hooks with the same name by
    creating individual hooks for any defined permutations
    of kwargs.

    This is accomplished by creating a unique name for each
    permutation proxying the original attribute name to the
    appropriate method determined from the provided kwargs.

    Args:
        name (str): Name of Proxy

    """

    def __init__(self, name: str):
        self.methods: List[ProxyItem[Callable[..., Any]]] = []
        self.instances: List[Type[ProjectModule]] = []
        self._name: str = name
        self.log: ServiceLog = Log.add_logger(str(self))

    def __call__(self, *args, **kwargs):
        proxy_kwargs = deepcopy(kwargs)
        proxy = self.resolve_proxy(**proxy_kwargs)
        if proxy:
            return getattr(*proxy)(*args, **kwargs)

    def __str__(self):
        name = f"HookProxy({self._name})"
        return name

    def __repr__(self):
        name = f"HookProxy(name={self._name}, methods=[{self.methods}])"
        return name

    def resolve_proxy(self, **kwargs: Any) -> (Type[ProjectModule], str):
        """Resolves appropriate instance and method to proxy to.

        If additional kwargs are provided and a proxy is not found,
        the function will continue to remove one kwarg and
        recurse into itself until either a match is found or it runs
        out of kwargs.

        Returns:
            Instance and method name if resolved, otherwise None.

        """
        proxy_kwargs = deepcopy(kwargs)
        for method, name in self.methods:
            _name = self.get_name(method, proxy_kwargs)
            if name == _name:
                instance = self._get_instance(method)
                self.log.debug(f"{self._name} proxied to [{_name}@{instance}]")
                return (instance, method.__name__)
        if proxy_kwargs:
            self.log.debug(
                f"could not resolve proxy: {self._name}[{proxy_kwargs}], broadening search..."
            )
            proxy_kwargs.popitem()
            return self.resolve_proxy(**proxy_kwargs)
        return None

    def _get_instance(self, attr: Callable[..., Any]) -> Optional[Type[ProjectModule]]:
        """Retrieves instance from attribute.

        Args:
            attr (Callable): Attribute to use.

        Returns:
            Instance the attribute belongs to.

        """
        _class = utils.get_class_that_defined_method(attr)
        if _class:
            instance = next((i for i in self.instances if isinstance(i, _class)), None)
            return instance

    def is_descriptor(self) -> bool:
        """Determine if initial method provided is a descriptor."""
        method = self.methods[0][0]
        instance = self._get_instance(method)
        if instance:
            attr = inspect.getattr_static(instance, self._name)
            return inspect.isdatadescriptor(attr)
        return False

    def get(self) -> T:
        """Get initial method descriptor value."""
        instance = self._get_instance(self.methods[0][0])
        self.log.debug(f"{self._name} proxied to [property@{instance}]")
        return getattr(instance, self._name)

    def add_method(self, func: Callable[..., Any], **kwargs: Any) -> ProxyItem[Callable[..., Any]]:
        """Adds method to Proxy.

        Any kwargs provided will be used to generate the unique
        hook name.

        Args:
            func (Callable): Method to add

        Example:
            >>> def test_func(arg1, kwarg1=False):
                    pass

            >>> self.add_method(test_func, {'kwarg1': False})
            (test_func, '_hook__test_func__kwarg1_False')

        Returns:
            Tuple[Callable, str]: Tuple containing method and unique hook name.

        """
        name = self.get_name(func, kwargs)
        hook = (func, name)
        self.methods.append(hook)
        self.log.debug(f"Method added to proxy: {hook}")
        return hook  # type: ignore

    def add_instance(self, inst: Any) -> Any:
        """Add instance to Proxy.

        Args:
            inst (Any): Instance to add.

        """
        return self.instances.append(inst)

    def get_name(self, func: Callable[..., Any], params: Optional[dict] = None) -> str:
        """Generates name from method and provided kwargs.

        Args:
            func (Callable): Method to generate name for.
            params (Dict[Any, Any], optional): Any kwargs to update the defaults with.
                Defaults to None. If none, uses default kwargs.

        Returns:
            str: Generated name

        """
        params = params or {}
        sig = inspect.signature(func)
        _default = {
            p.name: p.default
            for p in sig.parameters.values()
            if p.kind == p.POSITIONAL_OR_KEYWORD and p.default is not p.empty
        }
        params = {**_default, **params}
        name = f"_hook__{self._name}__{'__'.join(f'{k}_{v}' for k, v in params.items())}"
        return name



# -*- coding: utf-8 -*-

"""
micropy.stubs.source
~~~~~~~~~~~~~~

This module contains abstractions for handling stub sources
and their location.
"""


import io
import json
import shutil
import tarfile
import tempfile
from contextlib import contextmanager
from functools import partial
from pathlib import Path, PurePosixPath
from urllib import parse

import micropy.exceptions as exc
import requests
from micropy import utils
from micropy.logger import Log


class StubRepo:
    """Represents a remote repository for stubs.

    Args:
        name (str): Repo Name
        location (str): Valid url
        ref (str): path to repo definition file

    """

    repos = set()

    def __init__(self, name, location, path, **kwargs):
        self.name = name
        self.path = path
        self.log = Log.add_logger(self.name)
        self.location = location
        self.packages = kwargs.get("packages", [])
        self.repos.add(self)

    def has_package(self, name):
        """Checks if package is available in repo.

        Args:
            name (str): name of package

        Returns:
            bool: True if package is available

        """
        url = self.get_url(name)
        return utils.is_downloadable(url)

    def get_url(self, path):
        """Returns formatted url to provided path.

        Args:
            path (str): path to format

        Returns:
            str: formatted url

        """
        base_path = PurePosixPath(parse.urlparse(self.location).path)
        pkg_path = base_path / PurePosixPath(self.path) / PurePosixPath(path)
        url = parse.urljoin(self.location, str(pkg_path))
        self.log.debug(f"Stub Url: {url}")
        return url

    def search(self, query):
        """Searches repository packages.

        Args:
            query (str): query to search by

        Returns:
            [str]: List of matching results

        """
        query = query.strip().lower()
        pkg_names = [p["name"] for p in self.packages]
        results = set([p for p in pkg_names if query in p.lower()])
        return results

    @classmethod
    def resolve_package(cls, name):
        """Attempts to resolve package from all repos.

        Args:
            name (str): package to resolve

        Raises:
            StubNotFound: Package could not be resolved

        Returns:
            str: url to package

        """
        results = (r for r in cls.repos if r.has_package(name))
        try:
            repo = next(results)
        except StopIteration:
            raise exc.StubNotFound(name)
        else:
            pkg_url = repo.get_url(name)
            return pkg_url

    @classmethod
    def from_json(cls, content):
        """Create StubRepo Instances from JSON file.

        Args:
            file_obj (str or bytes): json content

        Returns:
            iterable of created repos

        """
        data = json.loads(content)
        for source in data:
            source_url = source["source"]
            source_data = requests.get(source_url).json()
            cls(**source_data)
        return cls.repos

    def __eq__(self, other):
        return self.location == getattr(other, "location", None)

    def __hash__(self):
        return hash(self.location)


class StubSource:
    """Abstract Base Class for Stub Sources."""

    def __init__(self, location, log=None):
        self.location = location
        _name = self.__class__.__name__
        self.log = log or Log.add_logger(_name)

    @contextmanager
    def ready(self, path=None, teardown=None):
        """Yields prepared Stub Source.

        Allows StubSource subclasses to have a preperation
        method before providing a local path to itself.

        Args:
            path (str, optional): path to stub source.
                Defaults to location.
            teardown (func, optional): callback to execute on exit.
                Defaults to None.

        Yields:
            Resolved PathLike object to stub source

        """
        _path = path or self.location
        info_path = next(_path.rglob("info.json"), None)
        path = Path(info_path.parent) if info_path else _path
        yield path
        if teardown:
            teardown()

    def __str__(self):
        _name = self.__class__.__name__
        return f"<{_name}@{self.location}>"


class LocalStubSource(StubSource):
    """Stub Source Subclass for local locations.

    Args:
        path (str): Path to Stub Source

    Returns:
        obj: Instance of LocalStubSource

    """

    def __init__(self, path, **kwargs):
        location = utils.ensure_existing_dir(path)
        return super().__init__(location, **kwargs)


class RemoteStubSource(StubSource):
    """Stub Source for remote locations.

    Args:
        url (str): URL to Stub Source

    Returns:
        obj: Instance of RemoteStubSource

    """

    def __init__(self, name, **kwargs):
        location = StubRepo.resolve_package(name)
        return super().__init__(location, **kwargs)

    def _unpack_archive(self, file_bytes, path):
        """Unpack archive from bytes buffer.

        Args:
            file_bytes (bytes): Byte array to extract from
                Must be from tarfile with gzip compression
            path (str): path to extract file to

        Returns:
            path: path extracted to

        """
        tar_bytes_obj = io.BytesIO(file_bytes)
        with tarfile.open(fileobj=tar_bytes_obj, mode="r:gz") as tar:
            tar.extractall(path)
        output = next(path.iterdir())
        return output

    def ready(self):
        """Retrieves and unpacks source.

        Prepares remote stub resource by downloading and
        unpacking it into a temporary directory.
        This directory is removed on exit of the superclass
        context manager

        Returns:
            callable: StubSource.ready parent method

        """
        tmp_dir = tempfile.mkdtemp()
        tmp_path = Path(tmp_dir)
        filename = utils.get_url_filename(self.location).split(".tar.gz")[0]
        _file_name = "".join(self.log.iter_formatted(f"$B[{filename}]"))
        content = utils.stream_download(
            self.location, desc=f"{self.log.get_service()} {_file_name}"
        )
        source_path = self._unpack_archive(content, tmp_path)
        teardown = partial(shutil.rmtree, tmp_path)
        return super().ready(path=source_path, teardown=teardown)


def get_source(location, **kwargs):
    """Factory for StubSource Instance.

    Args:
        location (str): PathLike object or valid URL

    Returns:
        obj: Either Local or Remote StubSource Instance

    """
    try:
        utils.ensure_existing_dir(location)
    except NotADirectoryError:
        return RemoteStubSource(location, **kwargs)
    else:
        return LocalStubSource(location, **kwargs)



# -*- coding: utf-8 -*-

from copy import deepcopy
from typing import Any, List, Sequence, Tuple, Type, Union

import dpath
from boltons import iterutils
from micropy.logger import Log, ServiceLog

from .config_json import JSONConfigSource
from .config_source import ConfigSource

"""Config Interface"""


class Config:
    """Configuration File Interface.

    Automatically syncs config in memory
    with config saved to disk.

    Args:
        path (Path): Path to save file at.
        source_format (ConfigSource, optional): Configuration File Format.
            Defaults to JSONConfigSource.
        default (dict, optional): Default configuration.
                Defaults to {}.

    """

    def __init__(
        self, *args: Any, source_format: Type[ConfigSource] = JSONConfigSource, default: dict = {}
    ):
        self.log: ServiceLog = Log.add_logger(f"{__name__}")
        self.format = source_format
        self._source: ConfigSource = self.format(*args)
        self._config = deepcopy(default)
        if self._source.exists:
            with self._source as src:
                self.log.debug("loaded config values")
                dpath.util.merge(self._config, src, flags=dpath.MERGE_REPLACE)

    @property
    def source(self) -> ConfigSource:
        return self._source

    @source.setter
    def source(self, value: Any) -> ConfigSource:
        self._source = self.format(value)
        return self._source

    @property
    def config(self) -> dict:
        return self._config

    def raw(self) -> dict:
        return self._config

    def sync(self) -> dict:
        """Sync in-memory config with disk.

        Returns:
            dict: updated config

        """
        with self.source as src:
            dpath.util.merge(src, self.config, flags=dpath.MERGE_REPLACE)
        return self.config

    def parse_key(self, key: str) -> Tuple[Sequence[str], str]:
        """Parses key.

        Splits it into a path and 'final key'
        object. Each key is seperates by a: "/"

        Example:
            >>> self.parse_key('item/subitem/value')
            (('item', 'subitem'), 'value')

        Args:
            key (str): key in dot notation

        Returns:
            Tuple[Sequence[str], str]: Parsed key

        """
        full_path = tuple(i for i in key.split("/"))
        path = full_path[:-1]
        p_key = full_path[-1]
        return (path, p_key)

    def get(self, key: str, default: Any = None) -> Any:
        """Retrieve config value.

        Args:
            key (str): Key (in dot-notation) of value to return.
            default (Any, optional): Default value to return.
                Defaults to None.

        Returns:
            Any: Value at key given

        """
        try:
            value = dpath.util.get(self.config, key)
        except KeyError:
            value = default
        else:
            return value
        return value

    def set(self, key: str, value: Any) -> Any:
        """Set config value.

        Args:
            key (str): Key (in dot-notation) to update.
            value (Any): Value to set

        Returns:
            Any: Updated config

        """
        dpath.set(self._config, key, value)
        self.log.debug(f"set config value [{key}] => {value}")
        return self.sync()

    def add(self, key: str, value: Any) -> Any:
        """Overwrite or add config value.

        Args:
            key: Key to set
            value: Value to add or update too

        Returns:
            Updated config

        """
        dpath.new(self._config, key, value)
        self.log.debug(f"added config value [{key}] -> {value}")
        return self.sync()

    def pop(self, key: str) -> Any:
        """Delete and return value at key.

        Args:
            key (str): Key to pop.

        Returns:
            Any: Popped value.

        """
        path, target = self.parse_key(key)
        value = self.get(key)
        remapped = iterutils.remap(
            self._config, lambda p, k, v: False if p == path and k == target else True
        )
        self._config = remapped
        self.log.debug(f"popped config value {value} <- [{key}]")
        return self.sync()

    def extend(self, key: str, value: List[Any], unique: bool = False) -> dict:
        """Extend a list in config at key path.

        Args:
            key: Key to path to extend.
            value: List of values to extend by.
            unique: Only extend values if not already in values.

        Returns:
            Updated Config

        """
        to_update = list(deepcopy(self.get(key, value)))
        if unique:
            value = [v for v in value if v not in to_update]
        dpath.merge(to_update, value, flags=dpath.MERGE_ADDITIVE)
        self.set(key, to_update)
        return self.sync()

    def upsert(self, key: str, value: Union[List[Any], dict]) -> dict:
        """Update or insert values into key list or dict.

        Args:
            key: Key to value to upsert.
            value: Value to upsert by.

        Returns:
            Updated config.

        """
        to_update = deepcopy(self.get(key, value))
        dpath.merge(to_update, value, flags=dpath.MERGE_REPLACE)
        self.add(key, to_update)
        return self.sync()

    def search(self, key):
        """Retrieve all values at key (with glob pattern).

        Args:
            key: Key with pattern to search with.

        Returns:
            Values matching key and pattern.

        """
        return dpath.values(self.config, key)



# -*- coding: utf-8 -*-

"""
micropy.stubs
~~~~~~~~~~~~~~

This module contains all functionality relating
to stub files/frozen modules and their usage in MicropyCli
"""

from . import source
from .stubs import StubManager

__all__ = ["StubManager", "source"]



"""Module for interfacing with py-devices."""

from .abc import (
    DevicePath,
    HostPath,
    MessageConsumer,
    MetaPyDevice,
    MetaPyDeviceBackend,
    PyDeviceConsumer,
    StreamConsumer,
)
from .consumers import ConsumerDelegate, MessageHandlers, ProgressStreamConsumer, StreamHandlers
from .pydevice import PyDevice

__all__ = [
    "PyDevice",
    "ConsumerDelegate",
    "ProgressStreamConsumer",
    "StreamHandlers",
    "MessageHandlers",
    "PyDeviceConsumer",
    "MessageConsumer",
    "StreamConsumer",
    "MetaPyDevice",
    "MetaPyDeviceBackend",
    "DevicePath",
    "HostPath",
]



# -*- coding: utf-8 -*-


from pathlib import Path
from typing import List, Optional, Tuple, Union

from .package import Package
from .source import DependencySource


class LocalDependencySource(DependencySource):
    """Dependency Source that is available locally.

    Args:
        package (Package): Package source points too.
        path (Path): Path to package.

    """

    def __init__(self, package: Package, path: Path):
        super().__init__(package)
        self._path = path
        self.is_local = True

    @property
    def path(self) -> Path:
        return self._path

    def __enter__(self) -> Union[Path, List[Tuple[Path, Optional[Path]]]]:
        """Determines appropriate path.

        Returns:
            Path to package root or list of files.

        """
        return self.path



# -*- coding: utf-8 -*-

import shutil

import pytest
from boltons import setutils
from micropy import config, packages, project
from micropy.exceptions import RequirementException
from micropy.project import modules
from requests import RequestException


@pytest.fixture
def get_module(tmp_path):
    def _get_module(names, mp, **kwargs):
        _templates = list(modules.TemplatesModule.TEMPLATES.keys())
        mods = {
            "stubs": (
                (
                    modules.StubsModule,
                    mp.stubs,
                ),
                {"stubs": list(mp.stubs)[:2]},
            ),
            "template": ((modules.TemplatesModule, _templates), {}),
            "reqs": ((modules.PackagesModule, "requirements.txt"), {}),
            "dev-reqs": ((modules.DevPackagesModule, "dev-requirements.txt"), {}),
        }
        if names == "all":
            names = ",".join(list(mods.keys()))
        _mods = [mods[n.strip()] for n in names.split(",") if n]
        for m in _mods:
            yield m

    return _get_module


@pytest.fixture
def get_config():
    def _get_config(request, name="NewProject", stubs=None, templates=None, packages={}):
        templates = templates or ["vscode", "pylint"]
        stubs = stubs or []
        _mods = {
            "base": {
                "name": name,
            },
            "stubs": {"stubs": {s.name: s.stub_version for s in stubs}},
            "template": {"config": {t: (t in templates) for t in templates}},
            "reqs": {"packages": packages.get("reqs", {})},
            "dev-reqs": {"dev-packages": packages.get("dev-reqs", {"micropy-cli": "*"})},
        }
        if request == "all":
            request = ",".join(list(_mods.keys()))
        mods = request.split(",")
        test_config = _mods["base"].copy()
        for m in mods:
            test_config = {**test_config, **_mods[m or "base"]}
        return test_config

    return _get_config


@pytest.fixture
def get_context():
    def _get_context(request, stubs=None, pkg_path=None, data_dir=None):
        stubs = stubs or []
        _frozen = [s.frozen for s in stubs]
        _fware = [s.firmware.frozen for s in stubs if s.firmware is not None]
        _stub_paths = [s.stubs for s in stubs]
        _paths = setutils.IndexedSet([*_frozen, *_fware, *_stub_paths])
        _context = {
            "base": {},
            "stubs": {
                "stubs": set(stubs),
                "paths": _paths,
                "datadir": data_dir,
            },
            "reqs": {"paths": setutils.IndexedSet([pkg_path]), "local_paths": set([])},
        }
        if request == "all":
            request = ",".join(list(_context.keys()))
        mods = request.split(",")
        if "reqs" in mods and "stubs" in mods:
            _ctx = _context["stubs"].copy()
            _ctx["paths"].update(_context["reqs"]["paths"])
            _ctx["local_paths"] = _context["reqs"]["local_paths"]
            return _ctx
        context = {}
        for m in mods:
            context = {**context, **_context.get(m, {})}
        return context

    return _get_context


@pytest.fixture
def test_project(micropy_stubs, tmp_path, get_module):
    def _test_project(mods="", path=None):
        mp = micropy_stubs()
        proj_path = path if path else tmp_path / "NewProject"
        proj = project.Project(proj_path)
        mods = get_module(mods, mp)
        for m in mods:
            proj.add(*m[0], **m[1])
        yield proj, mp

    return _test_project


@pytest.fixture
def tmp_project(tmp_path, shared_datadir):
    path = shared_datadir / "project_test"
    proj_path = tmp_path / "NewProject"
    shutil.copytree(path, proj_path)
    return proj_path


def test_implementation(mocker):
    """Test Abstract Base Class"""
    mocker.patch.object(modules.ProjectModule, "__abstractmethods__", new_callable=set)
    inst = modules.ProjectModule()
    inst.config
    inst.load()
    inst.create()
    inst.update()
    inst.add([])
    inst.remove([])


def test_project_queue(tmp_path, mock_cwd, mocker):
    mocker.patch("micropy.project.project.Config")
    path = tmp_path / "project_path"
    proj = project.Project(path)
    # should create/load projects based on <MODULES>.PRIORITY,
    # not based on order added
    mock_parent = mocker.Mock()
    mock_parent.m1.return_value = mocker.Mock(PRIORITY=8)
    mock_parent.m2.return_value = mocker.Mock(PRIORITY=0)
    mock_parent.m3.return_value = mocker.Mock(PRIORITY=9)
    proj.add(mock_parent.m1)
    proj.add(mock_parent.m2)
    proj.add(mock_parent.m3)
    proj.create()
    # should be called in order of priority (9 being highest)
    mock_parent.assert_has_calls(
        [
            # adding them doesnt matter
            mocker.call.m1(log=mocker.ANY, parent=mocker.ANY),
            mocker.call.m2(log=mocker.ANY, parent=mocker.ANY),
            mocker.call.m3(log=mocker.ANY, parent=mocker.ANY),
            # create order should differ
            mocker.call.m3().create(),
            mocker.call.m1().create(),
            mocker.call.m2().create(),
        ]
    )


@pytest.mark.parametrize("mods", ["", "stubs", "template", "reqs", "dev-reqs", "all"])
class TestProject:
    def test_create(self, test_project, mock_checks, mods, utils):
        test_proj, _ = next(test_project(mods))
        assert test_proj.config.get("name") == "NewProject"
        resp = test_proj.create()
        assert str(resp) == utils.str_path(test_proj.path)
        assert test_proj.exists
        assert test_proj.info_path.exists()
        if test_proj._children:
            test_proj.remove(type(test_proj._children[-1]))

    def test_config(self, test_project, get_config, mods, utils):
        test_proj, mp = next(test_project(mods))
        expect_config = get_config(mods, stubs=list(mp.stubs)[:2])
        assert test_proj.config._config == {"name": "NewProject"}
        test_proj.create()
        assert utils.dict_equal(test_proj.config.raw(), expect_config)
        # should be the same post-load
        test_proj.load()
        assert utils.dict_equal(test_proj.config.raw(), expect_config)

    def test_context(self, test_project, get_context, mods, tmp_path, utils):
        proj_path = tmp_path / "tmpprojpath"
        test_proj, mp = next(test_project(mods, path=proj_path))
        test_proj.create()
        pkg_path = test_proj.data_path / test_proj.name
        expect_context = get_context(
            mods, stubs=mp.stubs, pkg_path=pkg_path, data_dir=test_proj.data_path
        )
        assert utils.dict_equal(test_proj.context.raw(), expect_context)
        test_proj.load()  # should be the same post load
        assert utils.dict_equal(test_proj.context.raw(), expect_context)

    def test_load(self, mock_pkg, tmp_project, mock_checks, test_project, mods, utils):
        proj, mp = next(test_project(mods, path=tmp_project))
        proj.load(run_checks=mp.RUN_CHECKS)

    def test_update(self, mock_pkg, tmp_project, mock_checks, test_project, mods, utils):
        proj, mp = next(test_project(mods, path=tmp_project))
        proj.update()


class TestStubsModule:
    @pytest.fixture()
    def stub_module(self, mocker, tmp_path, micropy_stubs):
        mp = micropy_stubs()
        parent_mock = mocker.MagicMock()
        parent_mock.data_path = tmp_path / ".micropy"
        stub_item = list(mp.stubs)[0]
        stub_mod = modules.StubsModule(
            mp.stubs, stubs=[stub_item], parent=parent_mock, log=mocker.Mock()
        )
        return stub_mod, mp

    def test_load(self, tmp_project, stub_module, get_stub_paths):
        custom_stub = next(get_stub_paths())
        stub_mod, mp = stub_module
        stub_data = {
            "esp32-micropython-1.11.0": "1.2.0",
            "esp8266-micropython-1.11.0": "1.2.0",
            "custom-stub": str(custom_stub),
        }
        stub_mod.parent.config.get.return_value = stub_data
        stub_mod.stub_manager.add.return_value = mp.stubs
        assert stub_mod.load(stub_data=stub_data)

    def test_add_stub(self, test_project, get_stub_paths, mocker):
        proj, mp = next(test_project("stubs"))
        proj.create()
        mocker.patch.object(config.config.dpath, "merge")
        stub_path = next(get_stub_paths())
        stub = mocker.MagicMock()
        stub.path = stub_path
        stub.frozen = stub_path / "frozen"
        stub.stubs = stub_path / "stubs"
        stub.firmware = stub
        proj.add_stub(stub)
        print(proj.stubs)


class TestPackagesModule:
    @pytest.fixture
    def test_package(self, mocker, tmp_path, mock_pkg, test_project):
        new_path = tmp_path / "somepath"
        proj, mp = next(test_project("reqs", path=new_path))
        proj.create()
        return proj

    @pytest.mark.flaky
    def test_add_package(self, test_project, mock_pkg, tmp_path):
        proj, mp = next(test_project("reqs"))
        proj.create()
        proj.add_package("somepkg==7")
        # Shouldnt allow duplicate pkgs
        res = proj.add_package("somepkg")
        assert res is None

    @pytest.mark.flaky()
    def test_add_local_package(self, test_project, tmp_path, utils):
        proj, mp = next(test_project("reqs"))
        pkg = tmp_path / "custompkg"
        pkg.mkdir(parents=True)
        # Add from path
        proj.add_package(f"-e {pkg}")

    def test_package_error(self, test_project, mock_pkg, mocker, tmp_path, caplog):
        packages.source_package.utils.ensure_valid_url.side_effect = [RequestException]
        path = tmp_path / "newdir"
        proj, mp = next(test_project("reqs", path=path))
        proj.create()
        with pytest.raises(RequirementException):
            pkgs = proj.add_package("newpackage")
            assert proj.config.get("packages/newpackage", None) is None
            assert "newpackage" not in pkgs.keys()

    def test_add_dev_package(self, mocker, mock_pkg, test_project):
        proj, mp = next(test_project("reqs,dev-reqs"))
        proj.create()
        proj.add_package("somepkg")
        proj.add_package("anotha_pkg", dev=True)



# -*- coding: utf-8 -*-

"""
micropy.utils
~~~~~~~~~~~~~~

This module provides utility functions that are used within
MicropyCli.
"""

from .decorators import lazy_property
from .helpers import (
    create_dir_link,
    ensure_existing_dir,
    ensure_valid_url,
    extract_tarbytes,
    get_cached_data,
    get_class_that_defined_method,
    get_package_meta,
    get_url_filename,
    is_dir_link,
    is_downloadable,
    is_existing_dir,
    is_update_available,
    is_url,
    iter_requirements,
    search_xml,
    stream_download,
)
from .stub import generate_stub
from .validate import Validator



import sys

CRITICAL = 50
ERROR = 40
WARNING = 30
INFO = 20
DEBUG = 10
NOTSET = 0

_level_dict = {
    CRITICAL: "CRIT",
    ERROR: "ERROR",
    WARNING: "WARN",
    INFO: "INFO",
    DEBUG: "DEBUG",
}

_stream = sys.stderr


class Logger:

    level = NOTSET

    def __init__(self, name):
        self.name = name

    def _level_str(self, level):
        l = _level_dict.get(level)
        if l is not None:
            return l
        return "LVL%s" % level

    def setLevel(self, level):
        self.level = level

    def isEnabledFor(self, level):
        return level >= (self.level or _level)

    def log(self, level, msg, *args):
        if level >= (self.level or _level):
            _stream.write("%-6s:%-8s:" % (self._level_str(level), self.name))
            if not args:
                print(msg, file=_stream)
            else:
                print(msg % args, file=_stream)

    def debug(self, msg, *args):
        self.log(DEBUG, msg, *args)

    def info(self, msg, *args):
        self.log(INFO, msg, *args)

    def warning(self, msg, *args):
        self.log(WARNING, msg, *args)

    def error(self, msg, *args):
        self.log(ERROR, msg, *args)

    def critical(self, msg, *args):
        self.log(CRITICAL, msg, *args)

    def exc(self, e, msg, *args):
        self.log(ERROR, msg, *args)
        sys.print_exception(e, _stream)  # pylint: disable=no-member

    def exception(self, msg, *args):
        self.exc(sys.exc_info()[1], msg, *args)


_level = INFO
_loggers = {}


def getLogger(name):
    if name in _loggers:
        return _loggers[name]
    l = Logger(name)
    _loggers[name] = l
    return l


def info(msg, *args):
    getLogger(None).info(msg, *args)


def debug(msg, *args):
    getLogger(None).debug(msg, *args)


def basicConfig(level=INFO, filename=None, stream=None, format=None):
    global _level, _stream
    _level = level
    if stream:
        _stream = stream
    if filename is not None:
        print("logging.basicConfig: filename arg is not supported")
    if format is not None:
        print("logging.basicConfig: format arg is not supported")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['micropy',
 'micropy.config',
 'micropy.data',
 'micropy.lib.stubber',
 'micropy.packages',
 'micropy.project',
 'micropy.project.modules',
 'micropy.project.template.src',
 'micropy.pyd',
 'micropy.stubs',
 'micropy.utils']

package_data = \
{'': ['*'],
 'micropy.data': ['schemas/*'],
 'micropy.project': ['template/*', 'template/.vscode/*']}

install_requires = \
['GitPython>=3.1.7,<4.0.0',
 'Jinja2==3.0.3',
 'MarkupSafe==2.0.1',
 'boltons>=19.3,<20.0',
 'cachier>=1.2,<2.0',
 'click>=7.0,<8.0',
 'dpath>=1.4,<2.0',
 'jsonschema==3.2.0',
 'mypy>=0.940,<0.941',
 'packaging>=21.3,<22.0',
 'python-minifier>=2.5.0,<3.0.0',
 'questionary>=1.8.1,<2.0.0',
 'requests>=2.22,<3.0',
 'requirements-parser>=0.2.0,<0.3.0',
 'tqdm>=4.39,<5.0',
 'upydevice>=0.3.6,<0.4.0']

extras_require = \
{':python_version < "3.10" and sys_platform == "win32"': ['rshell>=0.0.31,<0.0.32'],
 ':sys_platform != "win32"': ['rshell>=0.0.31,<0.0.32'],
 ':sys_platform == "win32"': ['colorama>=0.4.3,<0.5.0']}

entry_points = \
{'console_scripts': ['micropy = micropy.cli:cli']}

setup_kwargs = {
    'name': 'micropy-cli',
    'version': '4.0.0c2',
    'description': 'Micropython Project Management Tool with VSCode support, Linting, Intellisense, Dependency Management, and more!',
    'long_description': '# Micropy Cli [![PyPI][pypi-img]][pypi-url] [![PyPI - Python Version][pypiv-img]][pypi-url] [![Github - Test Micropy Cli][build-img]][build-url] [![Coverage Status][cover-img]][cover-url]\n\n\nMicropy Cli is a project management/generation tool for writing [Micropython](https://micropython.org/) code in modern IDEs such as VSCode.\nIts primary goal is to automate the process of creating a workspace complete with:\n\n* **Linting** compatible with Micropython\n* VSCode **Intellisense**\n* **Autocompletion**\n* Dependency Management\n* VCS Compatibility\n\n\n<p align=\'center\'>\n    <img width=\'95%\' src=\'.github/img/micropy.svg\' alt="Micropy Demo SVG">\n</p>\n\n[pypi-img]: https://img.shields.io/pypi/v/micropy-cli?logo=pypi&logoColor=white&style=flat-square\n[pypi-url]: https://pypi.org/project/micropy-cli/\n[pypiv-img]: https://img.shields.io/pypi/pyversions/micropy-cli.svg?style=flat-square&logo=python&logoColor=green\n[build-img]: https://img.shields.io/github/workflow/status/BradenM/micropy-cli/Test%20MicropyCli/master?logo=github&style=flat-square\n[build-url]: https://github.com/BradenM/micropy-cli/actions\n[cover-img]: https://img.shields.io/coveralls/github/BradenM/micropy-cli/master?style=flat-square&logo=coveralls\n[cover-url]: https://coveralls.io/github/BradenM/micropy-cli\n\n# Getting Started\n\n## Installation\n\nYou can download and install the latest version of this software from the Python package index (PyPI) as follows:\n\n`pip install --upgrade micropy-cli`\n\nIf applicable, you can test out a pre-release by executing:\n\n`pip install --upgrade --pre micropy-cli`\n\n\n\n## Creating a Project\n\nCreating a new project folder is as simple as:\n\n1. Executing `micropy init <PROJECT NAME>`\n2. Selecting which features to enable\n3. Selecting your target device/firmware\n4. Boom. Your workspace is ready.\n\n<p align=\'center\'>\n    <img src=\'https://github.com/BradenM/micropy-cli/raw/master/.github/img/demo.gif\' alt="Micropy Demo">\n</p>\n\n\n## Micropy Project Environment\n\nWhen creating a project with `micropy-cli`, two special items are added:\n\n* A `.micropy/` folder\n* A `micropy.json` file\n\nThe `.micropy/` contains symlinks from your project to your `$HOME/.micropy/stubs` folder. By doing this, micropy can reference the required stub files for your project as relative to it, rather than using absolute paths to `$HOME/.micropy`. How does this benefit you? Thanks to this feature, you can feel free to push common setting files such as `settings.json` and `.pylint.rc` to your remote git repository. This way, others who clone your repo can achieve a matching workspace in their local environment.\n\n> Note: The generated `.micropy/` folder should be *IGNORED* by your VCS. It is created locally for each environment via the `micropy.json` file.\n\nThe `micropy.json` file contains information micropy needs in order to resolve your projects required files when other clone your repo. Think of it as a `package.json` for micropython.\n\n## Cloning a Micropy Environment\n\nTo setup a Micropy environment locally, simply:\n\n* Install `micropy-cli`\n* Navigate to the project directory\n* Execute `micropy`\n\nMicropy will automatically configure and install any stubs required by a project thanks to its `micropy.json` file.\n\n## Project Dependencies\n\nWhile all modules that are included in your targeted micropython firmware are available with autocompletion, intellisense, and linting, most projects require external dependencies.\n\nCurrently, handling dependencies with micropython is a bit tricky. Maybe you can install a cpython version of your requirement? Maybe you could just copy and paste it? What if it needs to be frozen?\n\nMicropy handles all these issues for you automatically. Not only does it track your project\'s dependencies, it keeps both `requirements.txt` and `dev-requirements.txt` updated, enables autocompletion/intellisense for each dep, and allows you to import them just as you would on your device.\n\nThis allows you to include your requirement however you want, whether that be as a frozen module in your custom built firmware, or simply in the `/lib` folder on your device.\n\n#### Installing Packages\n\nTo add a package as a requirement for your project, run:\n\n`micropy install <PACKAGE_NAMES>`\n\nwhile in your project\'s root directory.\n\nThis will automatically execute the following:\n\n* Source `PACKAGE_NAMES` from pypi, as a url, or a local path\n* Retrieve the module/package and stub it, adding it to your local `.micropy` folder.\n* Add requirement to your `micropy.json`\n* Update `requirements.txt`\n\nTo install dev packages that are not needed on your device, but are needed for local development, add the `--dev` flag. This will do everything above **except** stub the requirement.\n\nYou can also install all requirements found in `micropy.json`/`requirements.txt`/`dev-requirements.txt` by executing `micropy install` without passing any packages. Micropy will automatically do this when setting up a local environment of an existing micropy project.\n\n#### Example\n\nLets say your new project will depend on [picoweb](https://pypi.org/project/picoweb/) and [blynklib](https://pypi.org/project/blynklib/). Plus, you\'d like to use [rshell](https://pypi.org/project/rshell/) to communicate directly with your device. After creating your project via `micropy init`, you can install your requirements as so:\n\n<p align=\'center\'>\n    <img width="70%" src=\'.github/img/install_demo.svg\' alt="Micropy Pkg Install Demo">\n</p>\n\nNow you or anybody cloning your project can import those requirements normally, and have the benefits of all the features micropy brings:\n\n<p align=\'center\'>\n    <img width="70%" src=\'https://github.com/BradenM/micropy-cli/raw/master/.github/img/deps_demo.gif\' alt="Micropy Deps Demo">\n</p>\n\n\n## Stub Management\n\nStub files are the magic behind how micropy allows features such as linting, Intellisense, and autocompletion to work. To achieve the best results with MicropyCli, its important that you first add the appropriate stubs for the device/firmware your project uses.\n\n> Note: When working in a micropy project, all stub related commands will also be executed on the active project. (i.e if in a project and you run `micropy stubs add <stub-name>`, then that stub retrieved AND added to the active project.)\n\n### Adding Stubs\n\nAdding stubs to Micropy is a breeze. Simply run: `micropy stubs add <STUB_NAME>`\nBy sourcing [micropy-stubs](https://github.com/BradenM/micropy-stubs), MicroPy has several premade stub packages to choose from.\n\nThese packages generally use the following naming schema:\n\n`<device>-<firmware>-<version>`\n\nFor example, running `micropy stubs add esp32-micropython-1.11.0` will install the following:\n* Micropython Specific Stubs\n* ESP32 Micropython v1.11 Device Specific Stubs\n* Frozen Modules for both device and firmware\n\nYou can search stubs that are made available to Micropy via `micropy stubs search <QUERY>`\n\nAlternatively, using `micropy stubs add <PATH>`, you can manually add stubs to Micropy.\nFor manual stub generation, please see [Josvel/micropython-stubber](https://github.com/Josverl/micropython-stubber).\n\n### Creating Stubs\n\nUsing `micropy stubs create <PORT/IP_ADDRESS>`, MicropyCli can automatically generate and add stubs from any Micropython device you have on hand. This can be done over both USB and WiFi.\n\n> Note: For stub creation, micropy-cli has additional dependencies.\n>\n> These can be installed by executing: `pip install micropy-cli[create_stubs]`\n\n\n### Viewing Stubs\n\nTo list stubs you have installed, simply run `micropy stubs list`.\n\nTo search for stubs for your device, use `micropy stubs search <QUERY>`.\n\n# See Also\n\n* [VSCode IntelliSense, Autocompletion & Linting capabilities][lemariva-blog]\n    - An awesome article written by [lemariva](https://github.com/lemariva). It covers creating a micropython project environment from scratch using `micropy-cli` and [pymakr-vsc](pymakr-vsc). Great place to start if you\'re new to this!\n\n* [Developing for the Raspberry Pi Pico in VS Code][cpwood-medium]\n    - A getting started guide for developing in micropython on the Raspberry Pi Pico by [cpwood][cpwood-git].\n    - Also see: [Pico-Go: Micropy-Cli][cpwood-picogo]\n\n* [Awesome MicroPython][awesome-micropy]\n    - Collection of awesome micropython libraries / resources.\n    - Features `micropy-cli` along with several other great development tools under the [Development][awesome-micropy-develop] category.\n\n\n[lemariva-blog]: https://lemariva.com/blog/2019/08/micropython-vsc-ide-intellisense\n[lemariva-git]:  https://github.com/lemariva\n\n[cpwood-medium]: https://medium.com/all-geek-to-me/developing-for-the-raspberry-pi-pico-in-vs-code-getting-started-6dbb3da5ba97\n[cpwood-picogo]: http://pico-go.net/docs/help/micropy/\n[cpwood-git]: https://github.com/cpwood/\n\n[awesome-micropy]: https://awesome-micropython.com/\n[awesome-micropy-develop]: https://awesome-micropython.com/#development\n\n# Acknowledgements\n\n## Micropython-Stubber\n[Josvel/micropython-stubber](https://github.com/Josverl/micropython-stubber)\n\nJosverl\'s Repo is full of information regarding Micropython compatibility with VSCode and more. To find out more about how this process works, take a look at it.\n\nmicropy-cli and [micropy-stubs](https://github.com/BradenM/micropy-stubs) depend on micropython-stubber for its ability to generate frozen modules, create stubs on a pyboard, and more.\n',
    'author': 'Braden Mars',
    'author_email': 'bradenmars@bradenmars.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/BradenM/micropy-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



"""Micropy Exceptions."""


class MicropyException(Exception):
    """Generic MicroPy Exception."""


class StubError(MicropyException):
    """Exception for any errors raised by stubs."""

    def __init__(self, message=None, stub=None):
        super().__init__(message)
        self.stub = stub
        self.message = message
        if message is None:
            message = "An error occurred with this stub."


class StubValidationError(StubError):
    """Raised when a stub fails validation."""

    def __init__(self, path, errors, *args, **kwargs):
        msg = f"Stub at[{str(path)}] encountered" f" the following validation errors: {str(errors)}"
        super().__init__(message=msg, *args, **kwargs)

    def __str__(self):
        return self.message


class StubNotFound(StubError):
    """Raised when a stub cannot be found."""

    def __init__(self, stub_name=None):
        stub_name = stub_name or "Unknown"
        msg = f"{stub_name} is not available!"
        super().__init__(msg)


class RequirementException(MicropyException):
    """A Requirement Exception Occurred."""

    def __init__(self, *args, **kwargs):
        self.package = kwargs.pop("package", None)
        super().__init__(*args, **kwargs)


class RequirementNotFound(RequirementException):
    """A requirement could not be found."""


class PyDeviceError(MicropyException):
    """Generic PyDevice exception."""

    def __init__(self, message: str = None):
        super().__init__(message)
        self.message = message


class PyDeviceConnectionError(PyDeviceError):
    _default_message = "Failed to connect to pydevice @ {location}."

    def __init__(self, location: str):
        super().__init__(self._default_message.format(location=location))



import uos as os
import time


def countdown():
    for i in range(5, 0, -1):
        print("start stubbing in {}...".format(i))
        time.sleep(1)
    import createstubs


try:
    # only run import if no stubs yet
    os.listdir("stubs")
    print("stub folder was found, stubbing is not automatically started")

except OSError:
    countdown()
